#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Secrets Manager
# Copyright 2023 Keeper Security Inc.
# Contact: sm@keepersecurity.com
#

import argparse
import logging
import os
import sys
from keeper_secrets_manager_core.utils import bytes_to_base64, base64_to_bytes

from .base import Command, GroupCommand, dump_report_data, RecordMixin
from .tunnel.port_forward.TunnelGraph import TunnelDAG
from .tunnel.port_forward.tunnel_helpers import find_open_port, get_config_uid, get_keeper_tokens, \
    get_or_create_tube_registry, get_gateway_uid_from_record, resolve_record, resolve_pam_config, resolve_folder, \
    remove_field, start_rust_tunnel, get_tunnel_session, CloseConnectionReasons, create_rust_webrtc_settings
from .. import api, vault, record_management
from ..display import bcolors
from ..error import CommandError
from ..params import LAST_RECORD_UID
from ..subfolder import find_folders
from ..utils import value_to_boolean

# Group Commands
class PAMTunnelCommand(GroupCommand):

    def __init__(self):
        super(PAMTunnelCommand, self).__init__()
        self.register_command('start', PAMTunnelStartCommand(), 'Start Tunnel', 's')
        self.register_command('list', PAMTunnelListCommand(), 'List all Tunnels', 'l')
        self.register_command('stop', PAMTunnelStopCommand(), 'Stop Tunnel to the server', 'x')
        self.register_command('edit', PAMTunnelEditCommand(), 'Edit Tunnel settings', 'e')
        self.register_command('diagnose', PAMTunnelDiagnoseCommand(), 'Diagnose network connectivity to krelay server', 'd')
        self.default_verb = 'list'


class PAMConnectionCommand(GroupCommand):

    def __init__(self):
        super(PAMConnectionCommand, self).__init__()
        # self.register_command('start', PAMConnectionStartCommand(), 'Start Connection', 's')
        # self.register_command('stop', PAMConnectionStopCommand(), 'Stop Connection', 'x')
        self.register_command('edit', PAMConnectionEditCommand(), 'Edit Connection settings', 'e')
        self.default_verb = 'edit'


class PAMRbiCommand(GroupCommand):

    def __init__(self):
        super(PAMRbiCommand, self).__init__()
        self.register_command('edit', PAMRbiEditCommand(), 'Edit Remote Browser Isolation settings', 'e')
        self.default_verb = 'edit'


# Individual Commands
class PAMTunnelListCommand(Command):
    pam_cmd_parser = argparse.ArgumentParser(prog='pam tunnel list')

    def get_parser(self):
        return PAMTunnelListCommand.pam_cmd_parser

    def execute(self, params, **kwargs):
        # Try to get active tunnels from Rust PyTubeRegistry
        # Logger initialization is handled by get_or_create_tube_registry()
        tube_registry = get_or_create_tube_registry(params)
        if tube_registry:
            if not tube_registry.has_active_tubes():
                logging.warning(f"{bcolors.OKBLUE}No Tunnels running{bcolors.ENDC}")
                return

            table = []
            headers = ['Record', 'Remote Target', 'Local Address', 'Conversation ID', 'Status']

            # Get all tube IDs
            tube_ids = tube_registry.all_tube_ids()

            for tube_id in tube_ids:
                # Get conversation IDs for this tube
                conversation_ids = tube_registry.get_conversation_ids_by_tube_id(tube_id)

                # Get tunnel session for detailed info
                tunnel_session = get_tunnel_session(tube_id)

                # Record title
                record_title = tunnel_session.record_title if tunnel_session and tunnel_session.record_title else f"{bcolors.WARNING}unknown{bcolors.ENDC}"

                # Remote target
                if tunnel_session and tunnel_session.target_host and tunnel_session.target_port:
                    remote_target = f"{tunnel_session.target_host}:{tunnel_session.target_port}"
                else:
                    remote_target = f"{bcolors.WARNING}unknown{bcolors.ENDC}"

                # Local listening address
                if tunnel_session and tunnel_session.host and tunnel_session.port:
                    local_addr = f"{bcolors.OKGREEN}{tunnel_session.host}:{tunnel_session.port}{bcolors.ENDC}"
                else:
                    local_addr = f"{bcolors.WARNING}unknown{bcolors.ENDC}"

                # Conversation ID
                conv_id = conversation_ids[0] if conversation_ids else (tunnel_session.conversation_id if tunnel_session else 'none')

                # Connection state
                try:
                    state = tube_registry.get_connection_state(tube_id)
                    status_color = f"{bcolors.OKGREEN}" if state.lower() == "connected" else f"{bcolors.WARNING}"
                    status = f"{status_color}{state}{bcolors.ENDC}"
                except:
                    status = f"{bcolors.WARNING}unknown{bcolors.ENDC}"

                row = [
                    record_title,
                    remote_target,
                    local_addr,
                    conv_id,
                    status,
                ]
                table.append(row)

            dump_report_data(table, headers, fmt='table', filename="", row_number=False, column_width=None)
        else:
            # Rust WebRTC library is required for tunnel operations
            print(f"{bcolors.FAIL}This command requires the Rust WebRTC library (keeper_pam_webrtc_rs).{bcolors.ENDC}")
            print(f"{bcolors.OKBLUE}Please ensure the keeper_pam_webrtc_rs module is installed and available.{bcolors.ENDC}")
            return


class PAMTunnelStopCommand(Command):
    pam_cmd_parser = argparse.ArgumentParser(prog='pam tunnel stop')
    pam_cmd_parser.add_argument('uid', type=str, action='store', help='The Tunnel UID or Record UID')

    def get_parser(self):
        return PAMTunnelStopCommand.pam_cmd_parser

    def execute(self, params, **kwargs):
        uid = kwargs.get('uid')
        if not uid:
            raise CommandError('tunnel stop', '"uid" argument is required')

        tube_registry = get_or_create_tube_registry(params)
        if not tube_registry:
            raise CommandError('tunnel stop', 'This command requires the Rust WebRTC library')

        # Find matching tubes
        matching_tubes = tube_registry.find_tubes(uid)
        if not matching_tubes and tube_registry.tube_found(uid):
            matching_tubes = [uid]

        if not matching_tubes:
            raise CommandError('tunnel stop', f"No active tunnels found matching '{uid}'")

        # Close all matching tubes
        stopped_count = 0
        for tube_id in matching_tubes:
            try:
                tube_registry.close_tube(tube_id, reason=CloseConnectionReasons.Normal)
                print(f"{bcolors.OKGREEN}Stopped tunnel: {tube_id}{bcolors.ENDC}")
                stopped_count += 1
            except Exception as e:
                print(f"{bcolors.FAIL}Failed to stop tunnel {tube_id}: {e}{bcolors.ENDC}")

        if stopped_count == 0:
            raise CommandError('tunnel stop', f"Failed to stop any tunnels matching '{uid}'")


class PAMTunnelEditCommand(Command):
    pam_cmd_parser = argparse.ArgumentParser(prog='pam tunnel edit')
    pam_cmd_parser.add_argument('record', type=str, action='store', help='The record path or UID of the PAM '
                                                                      'resource record with network information to use '
                                                                      'for tunneling')
    pam_cmd_parser.add_argument('--configuration', '-c', required=False, dest='config', action='store',
                                help='The PAM Configuration UID or path to use for tunneling. '
                                     'Use command `pam config list` to view available PAM Configurations.')
    pam_cmd_parser.add_argument('--enable-tunneling', '-et', required=False, dest='enable_tunneling', action='store_true',
                                help='Enable tunneling on the record')
    pam_cmd_parser.add_argument('--tunneling-override-port', '-top', required=False, dest='tunneling_override_port',
                                action='store', help='Port to use for tunneling. If not provided, '
                                                     'the port from the record will be used.')
    pam_cmd_parser.add_argument('--disable-tunneling', '-dt', required=False, dest='disable_tunneling',
                                action='store_true', help='Disable tunneling on the record')
    pam_cmd_parser.add_argument('--remove-tunneling-override-port', '-rtop', required=False,
                                dest='remove_tunneling_override_port', action='store_true',
                                help='Remove tunneling override port')

    def get_parser(self):
        return PAMTunnelEditCommand.pam_cmd_parser

    def execute(self, params, **kwargs):
        tunneling_override_port = kwargs.get('tunneling_override_port')

        if ((kwargs.get('enable_tunneling') and kwargs.get('disable_tunneling')) or
                (kwargs.get('enable_rotation') and kwargs.get('disable_rotation')) or
                (kwargs.get('tunneling-override-port') and kwargs.get('remove_tunneling_override_port'))):
            raise CommandError('pam-config-edit', 'Cannot enable and disable the same feature at the same time')

        # First check if enabled is true, then check if disabled is true. If not, then set it to None
        _tunneling = True if kwargs.get('enable_tunneling') else False if  kwargs.get('disable_tunneling') else None
        _remove_tunneling_override_port = kwargs.get('remove_tunneling_override_port')

        if tunneling_override_port:
            try:
                tunneling_override_port = int(tunneling_override_port)
            except ValueError:
                raise CommandError('tunnel edit', 'tunneling-override-port must be an integer')

        record_name = kwargs.get('record')
        if not record_name:
            raise CommandError('pam tunnel edit', '"record" parameter is required.')
        record = RecordMixin.resolve_single_record(params, record_name)
        if not record:
            raise CommandError('pam tunnel edit', f'{bcolors.FAIL}Record \"{record_name}\" not found.{bcolors.ENDC}')
        if not isinstance(record, vault.TypedRecord):
            raise CommandError('pam tunnel edit', f'Record \"{record_name}\" can not be edited.')

        # config parameter is optional and maybe (auto)resolved from PAM record
        config_name = kwargs.get('config', None)
        cfg_rec = RecordMixin.resolve_single_record(params, config_name)
        if not cfg_rec and record.version == 6:
            cfg_rec = record  # trying to edit PAM Config itself
        config_uid = cfg_rec.record_uid if cfg_rec else None

        record_uid = record.record_uid
        record_type = record.record_type
        if record_type not in ("pamMachine pamDatabase pamDirectory pamNetworkConfiguration pamAwsConfiguration "
                               "pamRemoteBrowser pamAzureConfiguration").split():
            raise CommandError('', f"{bcolors.FAIL}This record's type is not supported for tunnels. "
                                   f"Tunnels are only supported on pamMachine, pamDatabase, pamDirectory, "
                                   f"pamRemoteBrowser, pamNetworkConfiguration pamAwsConfiguration, and "
                                   f"pamAzureConfiguration records{bcolors.ENDC}")

        encrypted_session_token, encrypted_transmission_key, transmission_key = get_keeper_tokens(params)
        if record_type in "pamNetworkConfiguration pamAwsConfiguration pamAzureConfiguration".split():
            tmp_dag = TunnelDAG(params, encrypted_session_token, encrypted_transmission_key, record_uid, is_config=True)
            tmp_dag.edit_tunneling_config(tunneling=_tunneling)
            tmp_dag.print_tunneling_config(record_uid, None)
        else:
            traffic_encryption_key = record.get_typed_field('trafficEncryptionSeed')
            # Generate a 256-bit (32-byte) random seed
            seed = os.urandom(32)
            dirty = False
            if not traffic_encryption_key or not traffic_encryption_key.value:
                base64_seed = bytes_to_base64(seed)
                record_seed = vault.TypedField.new_field('trafficEncryptionSeed', base64_seed, "")
                # if field is present update in-place, if in rec definition add to fields[] else custom[]
                record_types_with_seed = ("pamDatabase", "pamDirectory", "pamMachine", "pamRemoteBrowser")
                if traffic_encryption_key:
                    traffic_encryption_key.value = [base64_seed]
                elif record.get_record_type() in record_types_with_seed:
                    record.fields.append(record_seed)  # DU-469
                else:
                    record.custom.append(record_seed)
                dirty = True
            if dirty:
                record_management.update_record(params, record)
                api.sync_down(params)

                traffic_encryption_key = record.get_typed_field('trafficEncryptionSeed')
                if not traffic_encryption_key:
                    raise CommandError('', f"{bcolors.FAIL}Unable to add Seed to record {record_uid}. "
                                       f"Please make sure you have edit rights to record {record_uid} {bcolors.ENDC}")
            dirty = False

            existing_config_uid = get_config_uid(params, encrypted_session_token, encrypted_transmission_key, record_uid)

            tmp_dag = TunnelDAG(params, encrypted_session_token, encrypted_transmission_key, config_uid)
            old_dag = TunnelDAG(params, encrypted_session_token, encrypted_transmission_key, existing_config_uid)

            if config_uid and existing_config_uid != config_uid:
                old_dag.remove_from_dag(record_uid)
                tmp_dag.link_resource_to_config(record_uid)

            if tmp_dag is None or not tmp_dag.linking_dag.has_graph:
                raise CommandError('', f"{bcolors.FAIL}No PAM Configuration UID set. "
                                   f"This must be set or supplied for tunneling to work. This can be done by adding "
                                   f"{bcolors.OKBLUE}' --config [ConfigUID] "
                                   f" {bcolors.FAIL}The ConfigUID can be found by running "
                                   f"{bcolors.OKBLUE}'pam config list'{bcolors.ENDC}")

            if not tmp_dag.check_tunneling_enabled_config(enable_tunneling=_tunneling):
                tmp_dag.print_tunneling_config(config_uid, None)
                command = f"{bcolors.OKBLUE}'pam tunnel edit {config_uid}"
                if _tunneling and not tmp_dag.check_tunneling_enabled_config(
                        enable_tunneling=_tunneling):
                    command += f" --enable-tunneling" if _tunneling else ""

                print(f"{bcolors.FAIL}The settings are denied by PAM Configuration: {config_uid}. "
                      f"Please enable settings for the configuration by running\n"
                      f"{command}'{bcolors.ENDC}")
                return

            if not tmp_dag.is_tunneling_config_set_up(record_uid):
                tmp_dag.link_resource_to_config(record_uid)

            pam_settings = record.get_typed_field('pamSettings')
            if not pam_settings:
                pre_settings = {"connection": {}, "portForward": {}}
                if _tunneling and tunneling_override_port:
                    pre_settings["portForward"]["port"] = tunneling_override_port
                if pre_settings:
                    pam_settings = vault.TypedField.new_field('pamSettings', pre_settings, "")
                    record.custom.append(pam_settings)
                    dirty = True
            else:
                if not tmp_dag.is_tunneling_config_set_up(record_uid):
                    tmp_dag.link_resource_to_config(record_uid)
                if not pam_settings.value:
                    pam_settings.value.append({"connection": {}, "portForward": {}})
                if _tunneling and tunneling_override_port:
                    pam_settings.value[0]['portForward']['port'] = tunneling_override_port
                    dirty = True

                if _remove_tunneling_override_port and pam_settings.value[0]['portForward'].get('port'):
                    pam_settings.value[0]['portForward'].pop('port')
                    dirty = True
            if not tmp_dag.is_tunneling_config_set_up(record_uid):
                print(f"{bcolors.FAIL}No PAM Configuration UID set. This must be set for tunneling to work. "
                      f"This can be done by running "
                      f"{bcolors.OKBLUE}'pam tunnel edit {record_uid} --config [ConfigUID] --enable-tunneling' "
                      f"{bcolors.FAIL}The ConfigUID can be found by running "
                      f"{bcolors.OKBLUE}'pam config list'{bcolors.ENDC}")
                return
            allowed_settings_name = "allowedSettings"
            if record.record_type == "pamRemoteBrowser":
                allowed_settings_name = "pamRemoteBrowserSettings"

            if _tunneling is not None and tmp_dag.check_if_resource_allowed(record_uid, "portForwards") != _tunneling:
                dirty = True

            if dirty:
                tmp_dag.set_resource_allowed(resource_uid=record_uid, tunneling=_tunneling, allowed_settings_name=allowed_settings_name)

            # Print out the tunnel settings
            if not kwargs.get('silent'):
                tmp_dag.print_tunneling_config(record_uid, record.get_typed_field('pamSettings'), config_uid)


class PAMTunnelStartCommand(Command):
    pam_cmd_parser = argparse.ArgumentParser(prog='pam tunnel start')
    pam_cmd_parser.add_argument('uid', type=str, action='store', help='The Record UID of the PAM resource '
                                                                      'record with network information to use for '
                                                                      'tunneling')
    pam_cmd_parser.add_argument('--host', '-o', required=False, dest='host', action='store',
                                default="127.0.0.1",
                                help='The address on which the server will be accepting connections. It could be an '
                                     'IP address or a hostname. '
                                     'Ex. set to 127.0.0.1 as default so only connections from the same machine will be'
                                     ' accepted.')
    pam_cmd_parser.add_argument('--port', '-p', required=False, dest='port', action='store',
                                type=int, default=0,
                                help='The port number on which the server will be listening for incoming connections. '
                                     'If not set, random open port on the machine will be used.')
    pam_cmd_parser.add_argument('--no-trickle-ice', '-nti', required=False, dest='no_trickle_ice', action='store_true',
                                help='Disable trickle ICE for WebRTC connections. By default, trickle ICE is enabled '
                                     'for real-time candidate exchange.')

    def get_parser(self):
        return PAMTunnelStartCommand.pam_cmd_parser

    def execute(self, params, **kwargs):
        # Python version validation (same as before)
        from_version = [3, 8, 0]   # including
        major_version = sys.version_info.major
        minor_version = sys.version_info.minor
        micro_version = sys.version_info.micro

        if (major_version, minor_version, micro_version) < (from_version[0], from_version[1], from_version[2]):
            print(f"{bcolors.FAIL}This command requires Python {from_version[0]}.{from_version[1]}.{from_version[2]} or higher. "
                  f"You are using {major_version}.{minor_version}.{micro_version}.{bcolors.ENDC}")
            return

        # Check for Rust WebRTC library availability
        # Logger initialization is handled by get_or_create_tube_registry()
        tube_registry = get_or_create_tube_registry(params)
        if not tube_registry:
            print(f"{bcolors.FAIL}This command requires the Rust WebRTC library (keeper_pam_webrtc_rs).{bcolors.ENDC}")
            print(f"{bcolors.OKBLUE}Please ensure the keeper_pam_webrtc_rs module is installed and available.{bcolors.ENDC}")
            return

        record_uid = kwargs.get('uid')
        host = kwargs.get('host')
        port = kwargs.get('port')
        no_trickle_ice = kwargs.get('no_trickle_ice', False)

        if port is not None and port > 0:
            try:
                port = find_open_port(tried_ports=[], preferred_port=port, host=host)
            except CommandError as e:
                print(f"{bcolors.FAIL}{e}{bcolors.ENDC}")
                return
        else:
            port = find_open_port(tried_ports=[], host=host)
            if port is None:
                print(f"{bcolors.FAIL}Could not find open port to use for tunnel{bcolors.ENDC}")
                return

        # Sync and validate record
        api.sync_down(params)
        record = vault.KeeperRecord.load(params, record_uid)
        if not isinstance(record, vault.TypedRecord):
            print(f"{bcolors.FAIL}Record {record_uid} not found.{bcolors.ENDC}")
            return

        # Validate PAM settings
        pam_settings = record.get_typed_field('pamSettings')
        if not pam_settings:
            print(f"{bcolors.FAIL}PAM Settings not configured for record {record_uid}'.{bcolors.ENDC}")
            print(f"{bcolors.WARNING}This is done by running {bcolors.OKBLUE}'pam tunnel edit {record_uid} "
                  f"--enable-tunneling --config [ConfigUID]'"
                  f"{bcolors.WARNING} The ConfigUID can be found by running"
                  f"{bcolors.OKBLUE} 'pam config list'{bcolors.ENDC}.")
            return

        # Get target host and port
        target = record.get_typed_field('pamHostname')
        if not target:
            print(f"{bcolors.FAIL}Hostname not found for record {record_uid}.{bcolors.ENDC}")
            return
        target_host = target.get_default_value().get('hostName', None)
        target_port = target.get_default_value().get('port', None)
        if not target_host:
            print(f"{bcolors.FAIL}Host not found for record {record_uid}.{bcolors.ENDC}")
            return
        if not target_port:
            print(f"{bcolors.FAIL}Port not found for record {record_uid}.{bcolors.ENDC}")
            return

        # Check for SOCKS configuration
        allowed_hosts = record.get_typed_field('multiline', 'Allowed Hosts')
        allowed_ports = record.get_typed_field('multiline', 'Allowed Ports')
        socks = bool(allowed_hosts or allowed_ports)

        # Get encryption seed
        client_private_seed = record.get_typed_field('trafficEncryptionSeed')
        if not client_private_seed:
            print(f"{bcolors.FAIL}Traffic Encryption Seed not found for record {record_uid}.{bcolors.ENDC}")
            return
        base64_seed = client_private_seed.get_default_value(str).encode('utf-8')
        seed = base64_to_bytes(base64_seed)

        # Get gateway UID
        gateway_uid = get_gateway_uid_from_record(params, vault, record_uid)
        if not gateway_uid:
            print(f"{bcolors.FAIL}Gateway not found for record {record_uid}.{bcolors.ENDC}")
            return

        # Use Rust WebRTC implementation with configurable trickle ICE
        trickle_ice = not no_trickle_ice
        result = start_rust_tunnel(params, record_uid, gateway_uid, host, port, seed, target_host, target_port, socks, trickle_ice, record.title)
        
        if result and result.get("success"):
            # The helper will show endpoint table when local socket is actually listening
            pass
        else:
            # Print failure message
            error_msg = result.get("error", "Unknown error") if result else "Failed to start tunnel"
            fail_dynamic_length = len("| Tunnel failed to start: ") + len(error_msg)
            fail_dashed_line = '+' + '-' * fail_dynamic_length + '+'
            
            print(f'\n{bcolors.FAIL}{fail_dashed_line}{bcolors.ENDC}')
            print(f'{bcolors.FAIL}| Tunnel failed to start: {error_msg}{bcolors.ENDC}')
            print(f'{bcolors.FAIL}{fail_dashed_line}{bcolors.ENDC}\n')


class PAMTunnelDiagnoseCommand(Command):
    pam_cmd_parser = argparse.ArgumentParser(prog='pam tunnel diagnose', 
                                           description='Diagnose network connectivity to krelay server. '
                                                       'Tests DNS resolution, TCP/UDP connectivity, AWS infrastructure, '
                                                       'and WebRTC peer connection setup for IT troubleshooting.')
    pam_cmd_parser.add_argument('record', type=str, action='store', 
                                help='The Record UID of the PAM resource record to test connectivity for')
    pam_cmd_parser.add_argument('--timeout', '-t', required=False, dest='timeout', action='store',
                                type=int, default=30,
                                help='Test timeout in seconds (default: 30)')
    pam_cmd_parser.add_argument('--verbose', '-v', required=False, dest='verbose', action='store_true',
                                help='Show detailed diagnostic output including ICE server lists')
    pam_cmd_parser.add_argument('--format', '-f', required=False, dest='format', action='store',
                                choices=['table', 'json'], default='table',
                                help='Output format: table (human-readable) or json (machine-readable)')
    pam_cmd_parser.add_argument('--test', required=False, dest='test_filter', action='store',
                                help='Comma-separated list of specific tests to run. Available: '
                                     'dns_resolution,aws_connectivity,tcp_connectivity,udp_binding,'
                                     'ice_configuration,webrtc_peer_connection')

    def get_parser(self):
        return PAMTunnelDiagnoseCommand.pam_cmd_parser

    def execute(self, params, **kwargs):
        record_name = kwargs.get('record')
        timeout = kwargs.get('timeout', 30)
        verbose = kwargs.get('verbose', False)
        output_format = kwargs.get('format', 'table')
        test_filter = kwargs.get('test_filter')
        
        if not record_name:
            raise CommandError('pam tunnel diagnose', '"record" parameter is required.')

        # Check for Rust WebRTC library availability
        # Logger initialization is handled by get_or_create_tube_registry()
        tube_registry = get_or_create_tube_registry(params)
        if not tube_registry:
            print(f"{bcolors.FAIL}This command requires the Rust WebRTC library (keeper_pam_webrtc_rs).{bcolors.ENDC}")
            print(f"{bcolors.OKBLUE}Please ensure the keeper_pam_webrtc_rs module is installed and available.{bcolors.ENDC}")
            return 1

        # Resolve and validate the record
        api.sync_down(params)
        record = RecordMixin.resolve_single_record(params, record_name)
        if not record:
            print(f"{bcolors.FAIL}Record '{record_name}' not found.{bcolors.ENDC}")
            return 1
        if not isinstance(record, vault.TypedRecord):
            print(f"{bcolors.FAIL}Record '{record_name}' cannot be used for tunneling.{bcolors.ENDC}")
            return 1

        record_uid = record.record_uid
        record_type = record.record_type
        if record_type not in ("pamMachine pamDatabase pamDirectory pamNetworkConfiguration pamAwsConfiguration "
                               "pamRemoteBrowser pamAzureConfiguration").split():
            print(f"{bcolors.FAIL}Record type '{record_type}' is not supported for tunneling.{bcolors.ENDC}")
            print(f"Supported types: pamMachine, pamDatabase, pamDirectory, pamRemoteBrowser, "
                  f"pamNetworkConfiguration, pamAwsConfiguration, pamAzureConfiguration")
            return 1

        # Get the krelay server from the PAM configuration
        try:
            encrypted_session_token, encrypted_transmission_key, transmission_key = get_keeper_tokens(params)
            pam_config_uid = get_config_uid(params, encrypted_session_token, encrypted_transmission_key, record_uid)
            
            if not pam_config_uid:
                print(f"{bcolors.FAIL}No PAM Configuration found for record '{record_name}'.{bcolors.ENDC}")
                print(f"Please configure the record with: {bcolors.OKBLUE}pam tunnel edit {record_uid} --config [ConfigUID]{bcolors.ENDC}")
                return 1

            # The krelay server hostname is constructed from the params.server
            krelay_server = f"krelay.{params.server}"
            
        except Exception as e:
            print(f"{bcolors.FAIL}Failed to get PAM configuration: {e}{bcolors.ENDC}")
            return 1

        # Build test settings
        settings = {
            "use_turn": True,
            "turn_only": False
        }

        # Parse test filter if provided
        if test_filter:
            allowed_tests = {'dns_resolution', 'aws_connectivity', 'tcp_connectivity', 
                           'udp_binding', 'ice_configuration', 'webrtc_peer_connection'}
            requested_tests = set(test.strip() for test in test_filter.split(','))
            invalid_tests = requested_tests - allowed_tests
            if invalid_tests:
                print(f"{bcolors.FAIL}Invalid test names: {', '.join(invalid_tests)}{bcolors.ENDC}")
                print(f"Available tests: {', '.join(sorted(allowed_tests))}")
                return 1
            settings["test_filter"] = list(requested_tests)

        print(f"{bcolors.OKBLUE}Starting network connectivity diagnosis for krelay server: {krelay_server}{bcolors.ENDC}")
        print(f"Record: {record.title} ({record_uid})")
        print(f"Timeout: {timeout}s")
        print("")

        # Get TURN credentials for the connectivity test
        try:
            webrtc_settings = create_rust_webrtc_settings(
                params, host="127.0.0.1", port=0, 
                target_host="test", target_port=22, 
                socks=False, nonce=os.urandom(32)
            )
            turn_username = webrtc_settings.get("turn_username")
            turn_password = webrtc_settings.get("turn_password")
        except Exception as e:
            print(f"{bcolors.WARNING}Could not get TURN credentials: {e}{bcolors.ENDC}")
            turn_username = None
            turn_password = None

        # Run the connectivity test
        try:
            results = tube_registry.test_webrtc_connectivity(
                krelay_server=krelay_server,
                settings=settings,
                timeout_seconds=timeout,
                username=turn_username,
                password=turn_password
            )
            
            if output_format == 'json':
                import json
                print(json.dumps(results, indent=2))
                return 0
            else:
                # Use the built-in formatter for human-readable output
                formatted_output = tube_registry.format_connectivity_results(results, detailed=verbose)
                print(formatted_output)
                
                # Return appropriate exit code
                overall_result = results.get('overall_result', {})
                if overall_result.get('success', False):
                    return 0
                else:
                    return 1
                    
        except Exception as e:
            print(f"{bcolors.FAIL}Network connectivity test failed: {e}{bcolors.ENDC}")
            logging.debug(f"Full error details: {e}", exc_info=True)
            return 1


class PAMConnectionEditCommand(Command):
    choices = ['on', 'off', 'default']
    protocols = ['', 'http', 'kubernetes', 'mysql', 'postgresql', 'rdp', 'sql-server', 'ssh', 'telnet', 'vnc']
    parser = argparse.ArgumentParser(prog='pam connection edit')
    parser.add_argument('record', type=str, action='store', help='The record UID or path of the PAM '
                        'resource record with network information to use for connections')
    parser.add_argument('--configuration', '-c', required=False, dest='config', action='store',
                        help='The PAM Configuration UID or path to use for connections. '
                        'Use command `pam config list` to view available PAM Configurations.')
    parser.add_argument('--admin-user', '-a', required=False, dest='admin', action='store',
					help='The record path or UID of the PAM User record to configure the admin '
                    'credential on the PAM Resource')
    parser.add_argument('--protocol', '-p', dest='protocol', choices=protocols,
                        help='Set connection protocol')
    parser.add_argument('--connections', '-cn', dest='connections', choices=choices,
                        help='Set connections permissions')
    parser.add_argument('--connections-recording', '-cr', dest='recording', choices=choices,
                        help='Set recording connections permissions for the resource')
    parser.add_argument('--typescript-recording', '-tr', dest='typescriptrecording', choices=choices,
                        help='Set TypeScript recording permissions for the resource')
    parser.add_argument('--connections-override-port', '-cop', required=False, dest='connections_override_port',
                        action='store', help='Port to use for connections. If not provided, '
                        'the port from the record will be used.')
    parser.add_argument('--key-events', '-k', dest='key_events', choices=choices,
                        help='Toggle Key Events settings')
    parser.add_argument('--silent', '-s', required=False, dest='silent', action='store_true',
					help='Silent mode - don\'t print PAM User, PAM Config etc.')

    def get_parser(self):
        return PAMConnectionEditCommand.parser

    def execute(self, params, **kwargs):
        connection_override_port = kwargs.get('connections_override_port', None)

        # Convert on/off/default to True/False/None
        _connections = TunnelDAG._convert_allowed_setting(kwargs.get('connections', None))
        _recording = TunnelDAG._convert_allowed_setting(kwargs.get('recording', None))
        _typescript_recording = TunnelDAG._convert_allowed_setting(kwargs.get('typescriptrecording', None))

        if connection_override_port:
            try:
                connection_override_port = int(connection_override_port)
            except ValueError:
                raise CommandError('connection edit', '--connections-override-port must be an integer')

        record_name = kwargs.get('record')
        if not record_name:
            raise CommandError('pam connection edit', 'Record parameter is required.')
        record = RecordMixin.resolve_single_record(params, record_name)
        if not record:
            raise CommandError('pam connection edit', f'{bcolors.FAIL}Record \"{record_name}\" not found.{bcolors.ENDC}')
        if not isinstance(record, vault.TypedRecord):
            raise CommandError('pam connection edit', f'Record \"{record_name}\" can not be edited.')

        # config parameter is optional and maybe (auto)resolved from PAM record
        config_name = kwargs.get('config', None)
        cfg_rec = RecordMixin.resolve_single_record(params, config_name)
        if not cfg_rec and record.version == 6:
            cfg_rec = record  # trying to edit PAM Config itself
        config_uid = cfg_rec.record_uid if cfg_rec else None

        record_uid = record.record_uid
        record_type = record.record_type
        if record_type not in ("pamMachine pamDatabase pamDirectory pamNetworkConfiguration pamAwsConfiguration "
                               "pamRemoteBrowser pamAzureConfiguration").split():
            raise CommandError('', f"{bcolors.FAIL}This record's type is not supported for connections. "
                                   f"Connections are only supported on pamMachine, pamDatabase, pamDirectory, "
                                   f"pamRemoteBrowser, pamNetworkConfiguration pamAwsConfiguration, and "
                                   f"pamAzureConfiguration records{bcolors.ENDC}")

        encrypted_session_token, encrypted_transmission_key, transmission_key = get_keeper_tokens(params)
        if record_type in "pamNetworkConfiguration pamAwsConfiguration pamAzureConfiguration".split():
            tdag = TunnelDAG(params, encrypted_session_token, encrypted_transmission_key, record_uid, is_config=True)
            tdag.edit_tunneling_config(connections=_connections, session_recording=_recording, typescript_recording=_typescript_recording)
            if not kwargs.get("silent", False): tdag.print_tunneling_config(record_uid, None)
        else:
            traffic_encryption_key = record.get_typed_field('trafficEncryptionSeed')
            # Generate a 256-bit (32-byte) random seed
            seed = os.urandom(32)
            dirty = False
            if not traffic_encryption_key or not traffic_encryption_key.value:
                base64_seed = bytes_to_base64(seed)
                record_seed = vault.TypedField.new_field('trafficEncryptionSeed', base64_seed, "")
                # if field is present update in-place, if in rec definition add to fields[] else custom[]
                record_types_with_seed = ("pamDatabase", "pamDirectory", "pamMachine", "pamRemoteBrowser")
                if traffic_encryption_key:
                    traffic_encryption_key.value = [base64_seed]
                elif record.get_record_type() in record_types_with_seed:
                    record.fields.append(record_seed)  # DU-469
                else:
                    record.custom.append(record_seed)
                dirty = True

            protocol = kwargs.get("protocol", None)
            pam_settings = record.get_typed_field('pamSettings')
            if not pam_settings:
                pre_settings = {"connection": {}, "portForward": {}}
                if _connections:
                    if connection_override_port:
                        pre_settings["connection"]["port"] = connection_override_port
                    if protocol:
                        pre_settings["connection"]["protocol"] = protocol
                elif protocol or connection_override_port:
                    logging.warning(f'Connection override port and protocol can be set only when connections are enabled '
                            f'with {bcolors.OKGREEN}--connections=on{bcolors.ENDC} option')
                if pre_settings:
                    pam_settings = vault.TypedField.new_field('pamSettings', pre_settings, "")
                    record.custom.append(pam_settings)
                    dirty = True
            else:
                if not pam_settings.value:
                    pam_settings.value.append({"connection": {}, "portForward": {}})
                if not pam_settings.value[0]:
                    pam_settings.value[0] = {"connection": {}, "portForward": {}}
                if _connections:
                    if connection_override_port:
                        pam_settings.value[0]["connection"]["port"] = connection_override_port
                    elif connection_override_port is not None:  # empty string means remove port override
                        pam_settings.value[0]["connection"].pop("port", None)
                    if protocol:
                        pam_settings.value[0]["connection"]["protocol"] = protocol
                    elif protocol is not None:  # empty string means remove protocol
                        pam_settings.value[0]["connection"].pop("protocol", None)
                    dirty = True
                elif protocol or connection_override_port:
                    logging.warning(f'Connection override port and protocol can be set only when connections are enabled '
                            f'with {bcolors.OKGREEN}--connections=on{bcolors.ENDC} option')

            # pam_settings.value already initialized above
            key_events = kwargs.get('key_events')  # on/off/default
            if key_events:
                psv = pam_settings.value[0] if pam_settings and pam_settings.value else {}
                vcon = psv.get('connection', {}) if isinstance(psv, dict) else {}
                rik = vcon.get('recordingIncludeKeys') if isinstance(vcon, dict) else None
                if key_events == 'default':
                    if rik is not None:
                        pam_settings.value[0]["connection"].pop('recordingIncludeKeys', None)
                        dirty = True
                    else:
                        logging.debug(f'recordingIncludeKeys is already set to "default" on record={record_uid}')
                elif key_events == 'on':
                    if value_to_boolean(key_events) != value_to_boolean(rik):
                        pam_settings.value[0]["connection"]["recordingIncludeKeys"] = True
                        dirty = True
                    else:
                        logging.debug(f'recordingIncludeKeys is already enabled on record={record_uid}')
                elif key_events == 'off':
                    if value_to_boolean(key_events) != value_to_boolean(rik):
                        pam_settings.value[0]["connection"]["recordingIncludeKeys"] = False
                        dirty = True
                    else:
                        logging.debug(f'recordingIncludeKeys is already disabled on record={record_uid}')
                else:
                    logging.debug(f'Unexpected value for --key-events {key_events} (ignored)')

            if dirty:
                record_management.update_record(params, record)
                api.sync_down(params)

                traffic_encryption_key = record.get_typed_field('trafficEncryptionSeed')
                if not traffic_encryption_key:
                    raise CommandError('', f"{bcolors.FAIL}Unable to add Seed to record {record_uid}. "
                                       f"Please make sure you have edit rights to record {record_uid} {bcolors.ENDC}")
            dirty = False

            existing_config_uid = get_config_uid(params, encrypted_session_token, encrypted_transmission_key, record_uid)

            tdag = TunnelDAG(params, encrypted_session_token, encrypted_transmission_key, config_uid)
            old_dag = TunnelDAG(params, encrypted_session_token, encrypted_transmission_key, existing_config_uid)

            if config_uid and existing_config_uid != config_uid:
                old_dag.remove_from_dag(record_uid)
                tdag.link_resource_to_config(record_uid)

            if tdag is None or not tdag.linking_dag.has_graph:
                raise CommandError('', f"{bcolors.FAIL}No PAM Configuration UID set. "
                                   f"This must be set or supplied for connections to work. This can be done by adding "
                                   f"{bcolors.OKBLUE}' --config [ConfigUID] "
                                   f" {bcolors.FAIL}The ConfigUID can be found by running "
                                   f"{bcolors.OKBLUE}'pam config list'{bcolors.ENDC}")

            if not tdag.check_tunneling_enabled_config(enable_connections=_connections,
                                                       enable_session_recording=_recording,
                                                       enable_typescript_recording=_typescript_recording):
                if not kwargs.get("silent", False): tdag.print_tunneling_config(config_uid, None)
                command = f"{bcolors.OKBLUE}'pam connection edit {config_uid}"
                if _connections and not tdag.check_tunneling_enabled_config(enable_connections=_connections):
                    command += f" --connections=on" if _connections else ""
                if _recording and not tdag.check_tunneling_enabled_config(enable_session_recording=_recording):
                    command += f" --connections-recording=on" if _recording else ""
                if _typescript_recording and not tdag.check_tunneling_enabled_config(enable_typescript_recording=_typescript_recording):
                    command += f" --typescript-recording=on" if _typescript_recording else ""

                print(f"{bcolors.FAIL}The settings are denied by PAM Configuration: {config_uid}. "
                      f"Please enable settings for the configuration by running\n"
                      f"{command}'{bcolors.ENDC}")
                return

            if not tdag.is_tunneling_config_set_up(record_uid):
                tdag.link_resource_to_config(record_uid)

            if not tdag.is_tunneling_config_set_up(record_uid):
                print(f"{bcolors.FAIL}No PAM Configuration UID set. This must be set for connections to work. "
                      f"This can be done by running {bcolors.OKBLUE}"
                      f"'pam connection edit {record_uid} --config [ConfigUID] --enable-connections' "
                      f"{bcolors.FAIL}The ConfigUID can be found by running {bcolors.OKBLUE}'pam config list'{bcolors.ENDC}")
                return
            allowed_settings_name = "allowedSettings"
            if record.record_type == "pamRemoteBrowser":
                allowed_settings_name = "pamRemoteBrowserSettings"

            if _connections is not None and tdag.check_if_resource_allowed(record_uid, "connections") != _connections:
                dirty = True
            if _recording is not None and tdag.check_if_resource_allowed(record_uid, "sessionRecording") != _recording:
                dirty = True
            if _typescript_recording is not None and tdag.check_if_resource_allowed(record_uid, "typescriptRecording") != _typescript_recording:
                dirty = True

            if dirty:
                tdag.set_resource_allowed(resource_uid=record_uid,
                                          allowed_settings_name=allowed_settings_name,
                                          connections=kwargs.get('connections', None),
                                          session_recording=kwargs.get('recording', None),
                                          typescript_recording=kwargs.get('typescriptrecording', None))

            # admin parameter is optional yet if not set connections may fail
            admin_name = kwargs.get('admin')
            adm_rec = RecordMixin.resolve_single_record(params, admin_name)
            admin_uid = adm_rec.record_uid if adm_rec else None
            if admin_uid and record_type in ("pamDatabase", "pamDirectory", "pamMachine"):
                tdag.link_user_to_resource(admin_uid, record_uid, is_admin=True, belongs_to=True)
                # tdag.link_user_to_config(admin_uid)  # is_iam_user=True

            # Print out PAM Settings
            if not kwargs.get("silent", False): tdag.print_tunneling_config(record_uid, record.get_typed_field('pamSettings'), config_uid)

class PAMRbiEditCommand(Command):
    choices = ['on', 'off', 'default']
    parser = argparse.ArgumentParser(prog='pam rbi edit')
    parser.add_argument('--record', '-r', type=str, required=True, dest='record', action='store',
                        help='The record UID or path of the RBI record.')
    parser.add_argument('--configuration', '-c', required=False, dest='config', action='store',
                        help='The PAM Configuration UID or path to use for connections. '
                        'Use command `pam config list` to view available PAM Configurations.')
    parser.add_argument('--autofill-credentials', '-a', type=str, required=False, dest='autofill', action='store',
                        help='The record UID or path of the RBI Autofill Credentials record.')
    parser.add_argument('--key-events', '-k', dest='key_events', choices=choices,
                        help='Toggle Key Events settings')
    parser.add_argument('--remote-browser-isolation', '-rbi', dest='rbi', choices=choices,
                        help='Set RBI permissions')
    parser.add_argument('--connections-recording', '-cr', dest='recording', choices=choices,
                        help='Set recording connections permissions for the resource')
    parser.add_argument('--silent', '-s', required=False, dest='silent', action='store_true',
                        help='Silent mode - don\'t print PAM User, PAM Config etc.')

    def get_parser(self):
        return PAMRbiEditCommand.parser

    def execute(self, params, **kwargs):
        record_name = kwargs.get('record') or ''
        config_name = kwargs.get('config') or ''
        autofill = kwargs.get('autofill') or ''
        key_events = kwargs.get('key_events')  # on/off/default
        rbi = kwargs.get('rbi')  # on/off/default
        recording = kwargs.get('recording')  # on/off/default
        silent = kwargs.get('silent') or False

        if not record_name:
            raise CommandError('pam rbi edit', 'Record parameter is required.')
        if not (autofill or key_events or config_name or rbi or recording):
            raise CommandError('pam rbi edit', 'At least one parameter is required (-a -k -c -cr -rbi) '
                               ' and if the record is not linked to PAM Config -c option is required.')

        record = RecordMixin.resolve_single_record(params, record_name)
        if not record:
            raise CommandError('pam rbi edit', f'{bcolors.FAIL}Record \"{record_name}\" not found.{bcolors.ENDC}')
        if not isinstance(record, vault.TypedRecord):
            raise CommandError('pam rbi edit', f'Record \"{record_name}\" can not be edited.')

        record_uid = record.record_uid
        record_type = record.record_type
        if record_type != "pamRemoteBrowser":
            raise CommandError('pam rbi edit', f"{bcolors.FAIL}Record {record_uid} of type {record_type} "
                               "cannot be set up for RBI connections. "
                               f"RBI connection records must be of type: pamRemoteBrowser{bcolors.ENDC}")

        # record data (JSON) manipulations: autofill, key_events
        dirty = False
        traffic_encryption_key = record.get_typed_field('trafficEncryptionSeed')
        if not traffic_encryption_key or not traffic_encryption_key.value:
            seed = os.urandom(32)
            base64_seed = bytes_to_base64(seed)
            record_seed = vault.TypedField.new_field('trafficEncryptionSeed', base64_seed, "")
            if traffic_encryption_key:
                traffic_encryption_key.value = [base64_seed]
            else:
                record.fields.append(record_seed)
            dirty = True

        rbs_fld = record.get_typed_field('pamRemoteBrowserSettings')
        if not rbs_fld:
            rbsettings = {'connection': {'protocol': 'http', 'httpCredentialsUid': ''}}
            pam_rbsettings = vault.TypedField.new_field('pamRemoteBrowserSettings', rbsettings, '')
            record.fields.append(pam_rbsettings)
            dirty = True
        elif not rbs_fld.value:
            rbs_fld.value.append({'connection': {'protocol': 'http'}}) # type: ignore
            dirty = True

        if autofill:
            af_rec = RecordMixin.resolve_single_record(params, autofill)
            if not af_rec:
                raise CommandError('pam rbi edit', f'{bcolors.FAIL}Record \"{autofill}\" not found.{bcolors.ENDC}')
            if not isinstance(af_rec, vault.TypedRecord) or af_rec.version != 3 or af_rec.record_type not in ("login", "pamUser"):
                raise CommandError('pam rbi edit', f'Autofill credentials record \"{af_rec.record_uid}\" can not be linked. '
                                ' RBI autofill credential records must be of type "login" or "pamUser"')

            rbs_fld = record.get_typed_field('pamRemoteBrowserSettings')
            val1 = rbs_fld.value[0] if isinstance(rbs_fld, vault.TypedField) and rbs_fld.value else {}
            hcuid = val1.get('connection', {}).get('httpCredentialsUid') or '' if isinstance(val1, dict) else ''
            if af_rec.record_uid == hcuid:
                logging.debug(f'httpCredentialsUid={af_rec.record_uid} is already set up on record={record_uid}')
            elif rbs_fld and rbs_fld.value and isinstance(rbs_fld.value[0], dict):
                rbs_fld.value[0]["connection"]["httpCredentialsUid"] = af_rec.record_uid
                dirty = True
                if hcuid:
                    logging.debug(f'Updated existing httpCredentialsUid from: {hcuid} to: {af_rec.record_uid}')
            else:
                raise CommandError('pam rbi edit', f'{bcolors.FAIL}Failed to set httpCredentialsUid={af_rec.record_uid}{bcolors.ENDC}')

        if key_events:
            rbs_fld = record.get_typed_field('pamRemoteBrowserSettings')
            val1 = rbs_fld.value[0] if isinstance(rbs_fld, vault.TypedField) and rbs_fld.value else {}
            vcon = val1.get('connection', {}) if isinstance(val1, dict) else {}
            rik = vcon.get('recordingIncludeKeys') if isinstance(vcon, dict) else None
            if key_events == 'default':
                if rik is not None:
                    rbs_fld.value[0]["connection"].pop('recordingIncludeKeys', None)
                    dirty = True
                else:
                    logging.debug(f'recordingIncludeKeys is already set to "default" on record={record_uid}')
            elif key_events == 'on':
                if value_to_boolean(key_events) != value_to_boolean(rik):
                    rbs_fld.value[0]["connection"]["recordingIncludeKeys"] = True
                    dirty = True
                else:
                    logging.debug(f'recordingIncludeKeys is already enabled on record={record_uid}')
            elif key_events == 'off':
                if value_to_boolean(key_events) != value_to_boolean(rik):
                    rbs_fld.value[0]["connection"]["recordingIncludeKeys"] = False
                    dirty = True
                else:
                    logging.debug(f'recordingIncludeKeys is already disabled on record={record_uid}')
            else:
                logging.debug(f'Unexpected value for --key-events {key_events} (ignored)')

        if dirty:
            record_management.update_record(params, record)
            api.sync_down(params)

            traffic_encryption_key = record.get_typed_field('trafficEncryptionSeed')
            if not traffic_encryption_key:
                raise CommandError('', f"{bcolors.FAIL}Unable to add Seed to record {record_uid}. "
                                f"Please make sure you have edit rights to record {record_uid} {bcolors.ENDC}")
            params.sync_data = True

        # DAG manipulation options: config, rbi/connections, recording
        dirty = False
        if not (config_name or rbi or recording):
            return

        # resolve PAM Config
        encrypted_session_token, encrypted_transmission_key, _ = get_keeper_tokens(params)
        existing_config_uid = get_config_uid(params, encrypted_session_token, encrypted_transmission_key, record_uid)
        existing_config_uid = str(existing_config_uid) if existing_config_uid else ''

        # config parameter is optional and may be (auto)resolved from RBI record
        cfg_rec = None
        if config_name:
            cfg_rec = RecordMixin.resolve_single_record(params, config_name)
            msg = ("not found" if cfg_rec is None else "not the right type"
                   if not isinstance(cfg_rec, vault.TypedRecord) or cfg_rec.version != 6 else "")
            if msg:
                logging.warning(f'{bcolors.FAIL}PAM Config record "{config_name}" {msg} {bcolors.ENDC}')
                cfg_rec = None
        if not cfg_rec:
            logging.debug(f"PAM Config - using config from record {record_uid}")
            cfg_rec = RecordMixin.resolve_single_record(params, existing_config_uid)
            msg = ("not found" if cfg_rec is None else "not the right type"
                   if not isinstance(cfg_rec, vault.TypedRecord) or cfg_rec.version != 6 else "")
            if msg:
                logging.warning(f'{bcolors.FAIL}PAM Config record "{existing_config_uid}" {msg} {bcolors.ENDC}')
                cfg_rec = None

        config_uid = cfg_rec.record_uid if cfg_rec else None
        if not config_uid:
            raise CommandError('pam rbi edit', f'{bcolors.FAIL}PAM Config record not found.{bcolors.ENDC}')

        tdag = TunnelDAG(params, encrypted_session_token, encrypted_transmission_key, config_uid)
        if tdag is None or not tdag.linking_dag.has_graph:
            raise CommandError('', f"{bcolors.FAIL}No valid PAM Configuration UID set. "
                               "This must be set or supplied for connections to work. "
                               "The ConfigUID can be found by running "
                               f"{bcolors.OKBLUE}'pam config list'{bcolors.ENDC}")

        if config_uid:
            if existing_config_uid and existing_config_uid != config_uid:
                old_dag = TunnelDAG(params, encrypted_session_token, encrypted_transmission_key, existing_config_uid)
                old_dag.remove_from_dag(record_uid)
                logging.debug(f'Updated existing PAM Config UID from: {existing_config_uid} to: {config_uid}')
            tdag.link_resource_to_config(record_uid)

        # connections=on needed alongside remoteBrowserIsolation=on in PAM Config for RBI to work
        cfg_con_state = tdag.get_resource_setting(config_uid, 'allowedSettings', 'connections')
        cfg_rbi_state = tdag.get_resource_setting(config_uid, 'allowedSettings', 'remoteBrowserIsolation')
        cfg_rec_state = tdag.get_resource_setting(config_uid, 'allowedSettings', 'sessionRecording')
        if cfg_con_state != 'on' or cfg_rbi_state != 'on' or cfg_rec_state != 'on':
            if not silent:
                tdag.print_tunneling_config(config_uid, None)
            command = f"{bcolors.OKBLUE}'pam connection edit {config_uid}"
            command += ' --connections=on' if cfg_con_state != 'on' else ''
            command += ' --remote-browser-isolation=on' if cfg_rbi_state != 'on' else ''
            command += ' --connections-recording=on' if cfg_rec_state != 'on' else ''
            print(f"{bcolors.FAIL}Some settings may be denied by PAM Configuration: {config_uid} "
                  f" [ --connections={cfg_con_state} --remote-browser-isolation={cfg_rbi_state} "
                  f" --connections-recording={cfg_rec_state} ] "
                  f"To enable these settings for the configuration run\n"
                  f"{command}'{bcolors.ENDC}")

        if not tdag.is_tunneling_config_set_up(record_uid):
            tdag.link_resource_to_config(record_uid)

        if not tdag.is_tunneling_config_set_up(record_uid):
            print(f"{bcolors.FAIL}No PAM Configuration UID set. This must be set for connections to work. "
                f"This can be done by running {bcolors.OKBLUE}"
                f"'pam connection edit {record_uid} --config [ConfigUID] --enable-connections' "
                f"{bcolors.FAIL}The ConfigUID can be found by running {bcolors.OKBLUE}'pam config list'{bcolors.ENDC}")
            return

        con_val, rec_val = None, None
        rec_con_state = tdag.get_resource_setting(record_uid, 'allowedSettings', 'connections')
        rec_rec_state = tdag.get_resource_setting(record_uid, 'allowedSettings', 'sessionRecording')
        if (rbi is not None and rbi != rec_con_state) or (recording is not None and recording != rec_rec_state):
            con_val = rbi if rbi != rec_con_state else None
            rec_val = recording if recording != rec_rec_state else None
            dirty = True

        allowed_settings_name = "allowedSettings"
        # NB! Currently for remoteBrowserIsolation to work rec needs only "allowedSettings": {"connections": true}
        # allowed_settings_name = "pamRemoteBrowserSettings"
        # if rbi and rbi != tdag.get_resource_setting(record_uid, 'pamRemoteBrowserSettings', 'remoteBrowserIsolation'):
        #     dirty = True

        if dirty:
            tdag.set_resource_allowed(resource_uid=record_uid,
                                    allowed_settings_name=allowed_settings_name,
                                    connections=con_val,
                                    session_recording=rec_val)
        # if not kwargs.get("silent", False):
        #     tdag.print_tunneling_config(record_uid, record.get_typed_field('pamRemoteBrowserSettings'), config_uid)
        params.sync_data = True

class PAMSplitCommand(Command):
    pam_cmd_parser = argparse.ArgumentParser(prog='pam split')
    pam_cmd_parser.add_argument('pam_machine_record', type=str, action='store',
                                help='The record UID or title of the legacy PAM Machine '
                                'record with built-in PAM User credentials.')
    pam_cmd_parser.add_argument('--configuration', '-c', required=False, dest='pam_config', action='store',
                                help='The PAM Configuration Name or UID - If the legacy record was configured '
                                     'for rotation this command will try to autodetect PAM Configuration settings '
                                     'otherwise you\'ll be prompted to provide the PAM Config.')
    pam_cmd_parser.add_argument('--folder', '-f', required=False, dest='pam_user_folder', action='store',
                                help='The folder where to store the new PAM User record - '
                                     'folder names/paths are case sensitive!'
                                     '(if skipped - PAM User will be created into the '
                                     'same folder as PAM Machine)')

    def get_parser(self):
        return PAMSplitCommand.pam_cmd_parser

    def execute(self, params, **kwargs):
        # Parse command params
        pam_config = kwargs.get('pam_config', '')  # PAM Configuration Name or UID
        folder = kwargs.get('pam_user_folder', '')  # destination folder
        record_uid = kwargs.get('pam_machine_record', '')  # existing record UID

        record_uid = resolve_record(params, record_uid) or record_uid
        record = vault.KeeperRecord.load(params, record_uid)
        if not record:
            raise CommandError('', f"{bcolors.FAIL}Record {record_uid} not found.{bcolors.ENDC}")
        if not isinstance(record, vault.TypedRecord) or record.record_type != "pamMachine":
            raise CommandError('', f"{bcolors.FAIL}Record {record_uid} is not of the expected type 'pamMachine'.{bcolors.ENDC}")

        pam_config_uid = resolve_pam_config(params, record_uid, pam_config)
        if not pam_config_uid:
            print(f"{bcolors.FAIL}Please provide a valid PAM Configuration.{bcolors.ENDC}")
            return

        folder_uid = resolve_folder(params, folder)
        if folder and not folder_uid:
            print(f"{bcolors.WARNING}Unable to find destination folder '{folder}' "
                  "(Note: folder names/paths are case sensitive) "
                  "- PAM User record will be stored into same folder "
                  f"as the originating PAM Machine record.{bcolors.ENDC}")

        flogin = record.get_typed_field('login')
        vlogin = flogin.get_default_value(str) if flogin else ''
        fpass = record.get_typed_field('password')
        vpass = fpass.get_default_value(str) if fpass else ''
        fpkey = record.get_typed_field('secret')
        vpkey = fpkey.get_default_value(str) if fpkey else ''
        if not(vlogin or vpass or vpkey):
            if not(flogin or fpass or fpkey):
                print(f"{bcolors.WARNING}Record {record_uid} is already in the new format.{bcolors.ENDC}")
            else:
                # No values present - just drop the old fields and add new ones,
                # thus converting the record to the new pamMachine format
                # NB! If the record was edited - newer clients moved these to custom fields
                if flogin:
                    remove_field(record, flogin)
                if fpass:
                    remove_field(record, fpass)
                if fpkey:
                    remove_field(record, fpkey)

                if not record.get_typed_field('trafficEncryptionSeed'):
                    record_seed = vault.TypedField.new_field('trafficEncryptionSeed', "", "")
                    record.fields.append(record_seed)
                if not record.get_typed_field('pamSettings'):
                    pam_settings = vault.TypedField.new_field('pamSettings', "", "")
                    record.fields.append(pam_settings)

                record_management.update_record(params, record)
                params.sync_data = True

                print(f"{bcolors.WARNING}Record {record_uid} has no data to split and "
                    "was converted to the new format. Remember to manually add "
                    f"Administrative Credentials later.{bcolors.ENDC}")
            return
        elif not vlogin or not(vpass or vpkey):
            print(f"{bcolors.WARNING}Record {record_uid} has incomplete user data "
                  "but splitting anyway. Remember to manually update linked "
                  f"Administrative Credentials record later.{bcolors.ENDC}")

        # Create new pamUser record
        user_rec = vault.KeeperRecord.create(params, 'pamUser')
        user_rec.type_name = 'pamUser'
        user_rec.title = str(record.title) + ' Admin User'
        if flogin:
            field = user_rec.get_typed_field('login')
            field.value = flogin.value
        if fpass:
            field = user_rec.get_typed_field('password')
            field.value = fpass.value
        if fpkey:
            field = user_rec.get_typed_field('secret')
            field.value = fpkey.value

        if not folder_uid:  # use the folder of the PAM Machine record
            folders = list(find_folders(params, record.record_uid))
            uniq_items = len(set(folders))
            if uniq_items < 1:
                print(f"{bcolors.WARNING}The new record will be created in root folder.{bcolors.ENDC}")
            elif uniq_items > 1:
                print(f"{bcolors.FAIL}Record '{record.record_uid}' is probably "
                      "a linked record with copies/links across multiple folders "
                      f"and PAM User record will be created in folder '{folders[0]}'.{bcolors.ENDC}")
            folder_uid = folders[0] if folders else ''  # '' means root folder

        record_management.add_record_to_folder(params, user_rec, folder_uid)
        pam_user_uid = params.environment_variables.get(LAST_RECORD_UID, '')
        api.sync_down(params)

        if flogin:
            remove_field(record, flogin)
        if fpass:
            remove_field(record, fpass)
        if fpkey:
            remove_field(record, fpkey)

        if not record.get_typed_field('trafficEncryptionSeed'):
            record_seed = vault.TypedField.new_field('trafficEncryptionSeed', "", "")
            record.fields.append(record_seed)
        if not record.get_typed_field('pamSettings'):
            pam_settings = vault.TypedField.new_field('pamSettings', "", "")
            record.fields.append(pam_settings)

        record_management.update_record(params, record)
        params.sync_data = True

        if pam_config_uid:
            encrypted_session_token, encrypted_transmission_key, transmission_key = get_keeper_tokens(params)
            tdag = TunnelDAG(params, encrypted_session_token, encrypted_transmission_key, pam_config_uid, True)
            tdag.link_resource_to_config(record_uid)
            tdag.link_user_to_resource(pam_user_uid, record_uid, True, True)

        print(f"PAM Machine record {record_uid} user credentials were split into "
              f"a new PAM User record {pam_user_uid}")



import json
import logging
import os
from datetime import datetime

import requests
from keeper_secrets_manager_core.utils import bytes_to_base64, url_safe_str_to_bytes
from requests import ConnectionError
from urllib.parse import urlparse

import google

from . import gateway_helper
from .pam_dto import GatewayAction
from .. import base
from ... import crypto, utils, rest_api
from ...display import bcolors
from ...error import KeeperApiError
from ...params import KeeperParams
from ...proto import pam_pb2, router_pb2

VERIFY_SSL = bool(os.environ.get("VERIFY_SSL", "TRUE") == "TRUE")


def get_router_url(params: KeeperParams):
    krouter_env_var_name = "KROUTER_URL"
    if os.getenv(krouter_env_var_name):
        krouter_server_url = os.getenv(krouter_env_var_name)
        logging.debug(f"Getting Krouter url from ENV Variable '{krouter_env_var_name}'='{krouter_server_url}'")
    else:
        base_server_url = params.rest_context.server_base
        base_server = urlparse(base_server_url).netloc
        str_base_server = base_server
        if isinstance(base_server, bytes):
            base_server = base_server.decode('utf-8')
        
        # In GovCloud environments, the router service is not under the govcloud subdomain
        krouter_server_url = 'https://connect.' + base_server
        if '.govcloud.' in krouter_server_url:
            krouter_server_url = krouter_server_url.replace('.govcloud.', '.')

    return krouter_server_url


def get_router_ws_url(params: KeeperParams):
    router_url = get_router_url(params)
    # More precise replacement of just the scheme
    if router_url.startswith('https://'):
        router_url = 'wss://' + router_url[8:]
    elif router_url.startswith('http://'):
        router_url = 'ws://' + router_url[7:]
    return router_url


def router_get_connected_gateways(params):  # type: (KeeperParams) -> pam_pb2.PAMOnlineControllers
    rs = _post_request_to_router(params, 'get_controllers')

    if type(rs) == bytes:
        pam_online_controllers = pam_pb2.PAMOnlineControllers()
        pam_online_controllers.ParseFromString(rs)
        if logging.getLogger().level <= logging.DEBUG:
            js = google.protobuf.json_format.MessageToJson(pam_online_controllers)
            logging.debug('>>> [GW RS] %s: %s', 'get_controllers', js)

        return pam_online_controllers

    return None


# def router_get_record_rotation_info(params, record_uid_bytes):
#     rq = PAMGenericUidRequest()
#     rq.uid = record_uid_bytes
#
#     rs = _post_request_to_router(params, 'get_rotation_info', rq_proto=rq)
#
#     if type(rs) == bytes:
#         rri = RouterRotationInfo()
#         rri.ParseFromString(rs)
#
#         return rri
#
#     return None


def router_set_record_rotation_information(params, proto_request, transmission_key=None,
                                           encrypted_transmission_key=None, encrypted_session_token=None):
    rs = _post_request_to_router(params, 'set_record_rotation', proto_request, transmission_key=transmission_key,
                                 encrypted_transmission_key=encrypted_transmission_key,
                                 encrypted_session_token=encrypted_session_token)

    return rs


def router_get_rotation_schedules(params, proto_request):
    return _post_request_to_router(params, 'get_rotation_schedules', rq_proto=proto_request, rs_type=pam_pb2.PAMRotationSchedulesResponse)


def router_get_relay_access_creds(params, expire_sec=None):
    query_params = {
        'expire-sec': expire_sec
    }
    return _post_request_to_router(params, 'relay_access_creds', query_params=query_params, rs_type=pam_pb2.RelayAccessCreds)


def _post_request_to_router(params, path, rq_proto=None, rs_type=None, method='post',
                            raw_without_status_check_response=False, query_params=None, transmission_key=None,
                            encrypted_transmission_key=None, encrypted_session_token=None):
    krouter_host = get_router_url(params)
    path = '/api/user/' + path

    if not transmission_key:
        transmission_key = utils.generate_aes_key()
    if not encrypted_transmission_key:
        server_public_key = rest_api.SERVER_PUBLIC_KEYS[params.rest_context.server_key_id]

        if params.rest_context.server_key_id < 7:
            encrypted_transmission_key = crypto.encrypt_rsa(transmission_key, server_public_key)
        else:
            encrypted_transmission_key = crypto.encrypt_ec(transmission_key, server_public_key)

    encrypted_payload = b''

    if rq_proto:
        if logging.getLogger().level <= logging.DEBUG:
            js = google.protobuf.json_format.MessageToJson(rq_proto)
            logging.debug('>>> [GW RQ] %s: %s', path, js)
        encrypted_payload = crypto.encrypt_aes_v2(rq_proto.SerializeToString(), transmission_key)

    if not encrypted_session_token:
        encrypted_session_token = crypto.encrypt_aes_v2(utils.base64_url_decode(params.session_token), transmission_key)

    try:
        rs = requests.request(method,
                              krouter_host + path,
                              params=query_params,
                              verify=VERIFY_SSL,
                              headers={
                                'TransmissionKey': bytes_to_base64(encrypted_transmission_key),
                                'Authorization': f'KeeperUser {bytes_to_base64(encrypted_session_token)}'
                              },
                              data=encrypted_payload if rq_proto else None
        )
    except ConnectionError as e:
        raise KeeperApiError(-1, f"KRouter is not reachable on '{krouter_host}'. Error: ${e}")
    except Exception as ex:
        raise ex

    content_type = rs.headers.get('Content-Type') or ''

    if raw_without_status_check_response:
        return rs

    if rs.status_code < 400:
        if content_type == 'application/json':
            return rs.json()

        rs_body = rs.content
        if isinstance(rs_body, bytes):
            router_response = router_pb2.RouterResponse()
            router_response.ParseFromString(rs_body)

            rrc = router_pb2.RouterResponseCode.Name(router_response.responseCode)
            if router_response.responseCode != router_pb2.RRC_OK:
                raise Exception(router_response.errorMessage + ' Response code: ' + rrc)

            if router_response.encryptedPayload:
                payload_encrypted = router_response.encryptedPayload
                payload_decrypted = crypto.decrypt_aes_v2(payload_encrypted, transmission_key)
            else:
                payload_decrypted = None

            if rs_type:
                if payload_decrypted:
                    rs_proto = rs_type()
                    rs_proto.ParseFromString(payload_decrypted)
                    if logging.getLogger().level <= logging.DEBUG:
                        js = google.protobuf.json_format.MessageToJson(rs_proto)
                        logging.debug('>>> [GW RS] %s: %s', 'get_rotation_schedules', js)
                    return rs_proto
                else:
                    return None

            return payload_decrypted

        return rs_body
    else:
        raise KeeperApiError(rs.status_code, rs.text)


def get_controller_cookie(params, destination_controller_uid_str):

    # TODO: Cache the cookies for controller UIDs to improve the performance
    max_count = 100
    curr_count = 0

    while True:
        if curr_count > max_count:
            logging.error(f"Too many calls without getting good response from the server. max_count={max_count}")

            return None

        resp = _post_request_to_router(params,
                                       f'bind_to_controller/{destination_controller_uid_str}',
                                       method='get',
                                       raw_without_status_check_response=True)

        # print('Cookies:')
        # for c in resp.cookies:
        #     print(c.name, c.value)

        if resp.status_code == 200:
            logging.debug("Found right host")
            return resp.cookies
        if resp.status_code == 303:
            logging.debug("Controller connected to the router, but on another host. Try another call...")
        else:
            logging.warning("Looks like there is no such controller connected to the router.")
            return None

def request_cookie_jar_to_str(cookie_jar):
    cookie_dict = dict(cookie_jar)
    found = ['%s=%s' % (name, value) for (name, value) in cookie_dict.items()]
    return ';'.join(found)


def router_send_action_to_gateway(params, gateway_action: GatewayAction, message_type, is_streaming,
                                  destination_gateway_uid_str=None, gateway_timeout=15000, transmission_key=None,
                                  encrypted_transmission_key=None, encrypted_session_token=None,
                                  destination_gateway_cookies=None):
    # Default time out how long the response from the Gateway should be
    krouter_host = get_router_url(params)

    # 1. Find connected gateway to send action to
    try:
        router_enterprise_controllers_connected = \
            [x.controllerUid for x in router_get_connected_gateways(params).controllers]

    except requests.exceptions.ConnectionError as errc:
        logging.info(f"{bcolors.WARNING}Looks like router is down. Router URL [{krouter_host}]{bcolors.ENDC}")
        return
    except Exception as e:
        raise e

    if destination_gateway_uid_str:
        # Means that we want to get info for a specific Gateway

        destination_gateway_uid_bytes = url_safe_str_to_bytes(destination_gateway_uid_str)

        if destination_gateway_uid_bytes not in router_enterprise_controllers_connected:
            print(f"{bcolors.WARNING}\tThis Gateway currently is not online.{bcolors.ENDC}")
            return
    else:
        if not router_enterprise_controllers_connected or len(router_enterprise_controllers_connected) == 0:
            print(f"{bcolors.WARNING}\tNo running or connected Gateways in your enterprise. "
                  f"Start the Gateway before sending any action to it.{bcolors.ENDC}")
            return
        elif len(router_enterprise_controllers_connected) == 1:
            destination_gateway_uid_bytes = router_enterprise_controllers_connected[0]
            destination_gateway_uid_str = utils.base64_url_encode(destination_gateway_uid_bytes)
        else:  # There are more than two Gateways connected. Selecting the right one

            if not gateway_action.gateway_destination:
                print(f"{bcolors.WARNING}There are more than one Gateways running in your enterprise. "
                      f"Only '{bcolors.OKGREEN}pam action rotate{bcolors.WARNING}' is able to know "
                      f"which Gateway should receive a request. Any other commands should have a Gateway specified. "
                      f"See help for the command you are trying to use. To find connected gateways run action "
                      f"'{bcolors.OKGREEN}pam gateway list{bcolors.WARNING}' and provide Gateway UID or Gateway Name.{bcolors.ENDC}")

                return

            destination_gateway_uid_bytes = gateway_helper.find_connected_gateways(router_enterprise_controllers_connected, gateway_action.gateway_destination)
            destination_gateway_uid_str = utils.base64_url_encode(destination_gateway_uid_bytes)

    msg_id = gateway_action.conversationId if gateway_action.conversationId else GatewayAction.generate_conversation_id('true')

    rq = router_pb2.RouterControllerMessage()
    rq.messageUid = utils.base64_url_decode(msg_id) if isinstance(msg_id, str) else msg_id
    rq.controllerUid = destination_gateway_uid_bytes
    rq.messageType = message_type
    rq.streamResponse = is_streaming
    rq.payload = gateway_action.toJSON().encode('utf-8')
    rq.timeout = gateway_timeout

    if not transmission_key:
        transmission_key = utils.generate_aes_key()

    response = router_send_message_to_gateway(
        params=params,
        transmission_key=transmission_key,
        rq_proto=rq,
        destination_gateway_uid_str=destination_gateway_uid_str,
        destination_gateway_cookies=destination_gateway_cookies,
        encrypted_transmission_key=encrypted_transmission_key,
        encrypted_session_token=encrypted_session_token)

    rs_body = response.content

    if type(rs_body) == bytes:
        router_response = router_pb2.RouterResponse()
        router_response.ParseFromString(rs_body)

        rrc = router_pb2.RouterResponseCode.Name(router_response.responseCode)
        if router_response.responseCode == router_pb2.RRC_OK:
            logging.debug("Good response...")

        elif router_response.responseCode == router_pb2.RRC_BAD_STATE:
            raise Exception(router_response.errorMessage + ' response code: ' + rrc)

        elif router_response.responseCode == router_pb2.RRC_TIMEOUT:
            # Router tried to send message to the Controller but the response didn't arrive on time
            # ex. if Router is expecting response to be within 3 sec, but the gateway didn't respond within that time
            raise Exception(router_response.errorMessage + ' response code: ' + rrc)

        elif router_response.responseCode == router_pb2.RRC_CONTROLLER_DOWN:
            # Sent an action to the Controller that is no longer online
            raise Exception(router_response.errorMessage + ' response code: ' + rrc)

        else:
            raise Exception(router_response.errorMessage + ' response code: ' + rrc)


        payload_encrypted = router_response.encryptedPayload
        if payload_encrypted:

            payload_decrypted = crypto.decrypt_aes_v2(payload_encrypted, transmission_key)

            controller_response = pam_pb2.ControllerResponse()
            controller_response.ParseFromString(payload_decrypted)

            gateway_response_payload = json.loads(controller_response.payload)
        else:
            gateway_response_payload = {}

        # controller_response_str = router_response.get('data')
        # controller_response = json.loads(controller_response_str)

        return {
            'response': gateway_response_payload
        }


def router_send_message_to_gateway(params, transmission_key, rq_proto, destination_gateway_uid_str,
                                   destination_gateway_cookies=None, encrypted_transmission_key=None,
                                   encrypted_session_token=None):

    krouter_host = get_router_url(params)

    if not encrypted_transmission_key:
        server_public_key = rest_api.SERVER_PUBLIC_KEYS[params.rest_context.server_key_id]

        if params.rest_context.server_key_id < 7:
            encrypted_transmission_key = crypto.encrypt_rsa(transmission_key, server_public_key)
        else:
            encrypted_transmission_key = crypto.encrypt_ec(transmission_key, server_public_key)

    encrypted_payload = b''

    if rq_proto:
        encrypted_payload = crypto.encrypt_aes_v2(rq_proto.SerializeToString(), transmission_key)
    if not encrypted_session_token:
        encrypted_session_token = crypto.encrypt_aes_v2(utils.base64_url_decode(params.session_token), transmission_key)

    if not destination_gateway_cookies:
        destination_gateway_cookies = get_controller_cookie(params, destination_gateway_uid_str)

    if not destination_gateway_cookies:
        raise Exception('Even though it seems that the Gateway is online, Commander was unable to get the '
                        'cookies to connect to the Gateway')

    rs = requests.post(
        krouter_host+"/api/user/send_controller_message",
        verify=VERIFY_SSL,

        headers={
            'TransmissionKey': bytes_to_base64(encrypted_transmission_key),
            'Authorization': f'KeeperUser {bytes_to_base64(encrypted_session_token)}',
        },
        cookies=destination_gateway_cookies,
        data=encrypted_payload if rq_proto else None
    )

    if rs.status_code >= 300:
        raise Exception(str(rs.status_code) + ': error: ' + rs.reason + ', message: ' + rs.text)

    return rs


def get_response_payload(router_response):

    router_response_response = router_response.get('response')
    router_response_response_payload_str = router_response_response.get('payload')
    router_response_response_payload_dict = json.loads(router_response_response_payload_str)

    return router_response_response_payload_dict


def get_dag_leafs(params, encrypted_session_token, encrypted_transmission_key, record_id):
    """
    POST a stringified JSON object to /api/dag/get_leafs on the KRouter
    The object is:
    {
      vertex: string,
      graphId: number
    }
    """
    krouter_host = get_router_url(params)
    path = '/api/user/get_leafs'

    payload = {
        'vertex': record_id,
        'graphId': 0
    }

    try:
        rs = requests.request('post',
                              krouter_host + path,
                              verify=VERIFY_SSL,
                              headers={
                                  'TransmissionKey': bytes_to_base64(encrypted_transmission_key),
                                  'Authorization': f'KeeperUser {bytes_to_base64(encrypted_session_token)}'
                              },
                              data=json.dumps(payload).encode('utf-8')
                              )
    except ConnectionError as e:
        raise KeeperApiError(-1, f"KRouter is not reachable on '{krouter_host}'. Error: ${e}")
    except Exception as ex:
        raise ex

    if rs.status_code == 200:
        logging.debug("Found right host")
        return rs.json()
    else:
        logging.warning("Looks like there is no such controller connected to the router.")
        return None


def print_router_response(router_response, response_type, original_conversation_id=None, is_verbose=False, gateway_uid=''):
    if not router_response:
        return

    router_response_response = router_response.get('response')
    router_response_response_payload_str = router_response_response.get('payload')
    router_response_response_payload_dict = json.loads(router_response_response_payload_str)

    if router_response_response_payload_dict.get('warnings'):
        for w in router_response_response_payload_dict.get('warnings'):
            if w:
                print(f'{bcolors.WARNING}{w}{bcolors.ENDC}')

    if original_conversation_id:
        # gateway_response_conversation_id = utils.base64_url_decode(router_response_response_payload_dict.get('conversation_id')).decode("utf-8")
        # IDs are either bytes or base64 encoded strings which may be padded
        gateway_response_conversation_id = router_response_response_payload_dict.get('conversation_id', None)
        oid = (utils.base64_url_decode(original_conversation_id)
               if isinstance(original_conversation_id, str)
               else original_conversation_id)
        gid = (utils.base64_url_decode(gateway_response_conversation_id)
               if isinstance(gateway_response_conversation_id, str)
               else gateway_response_conversation_id)

        if oid != gid:
            logging.error(f"Message ID that was sent to the server [{original_conversation_id}] and the conversation id "
                          f"received back [{gateway_response_conversation_id}] are different. That probably means that "
                          f"the gateway sent a wrong response that was not associated with the request.")

    if not (router_response_response_payload_dict.get('is_ok') or router_response_response_payload_dict.get('isOk')):
        print(f"{bcolors.FAIL}{json.dumps(router_response_response_payload_dict, indent=4)}{bcolors.ENDC}")
        return

    if router_response_response_payload_dict.get('isScheduled') or router_response_response_payload_dict.get('is_scheduled'):
        conversation_id = router_response_response_payload_dict.get('conversation_id')

        gwinfo = f" --gateway={gateway_uid}" if gateway_uid else ""
        print(f"Scheduled action id: {bcolors.OKBLUE}{conversation_id}{bcolors.ENDC}")
        print(f"The action has been scheduled, use command '{bcolors.OKGREEN}pam action job-info {conversation_id}{gwinfo}{bcolors.ENDC}' to get status of the scheduled action")
        return

    elif response_type == 'job_info':
        job_info = router_response_response_payload_dict.get('data')
        exec_response_value = job_info.get('execResponseValue')
        exec_response_value_msg = exec_response_value.get('message') if exec_response_value else None
        exec_response_value_logs = exec_response_value.get('execLog') if exec_response_value else None
        exec_duration = job_info.get('executionDuration')
        exec_status = job_info.get('status')
        exec_exception = job_info.get('execException')

        print(f'Execution Details\n-------------------------')

        if exec_status == 'finished':
            font_color = bcolors.OKGREEN
        elif exec_status == 'running':
            font_color = bcolors.WARNING
        else:
            font_color = bcolors.FAIL

        print(f'\t{font_color}Status              : {job_info.get("reason") if job_info.get("reason") else exec_status}{bcolors.ENDC}')

        if exec_duration:
            print(f'\t{font_color}Duration            : {exec_duration}{bcolors.ENDC}')

        if exec_response_value_msg:
            print(f'\t{font_color}Response Message    : {exec_response_value_msg}{bcolors.ENDC}')

        if exec_response_value_logs:
            print(f'\t{font_color}Post-execution scripts logs:{bcolors.ENDC}')
            for el in exec_response_value_logs:
                print(f'\t\t{font_color}script: {el.get("name")}{bcolors.ENDC}')
                print(f'\t\t{font_color}return code: {el.get("return_code")}{bcolors.ENDC}')
                if el.get("stdout"):
                    print(f'\t\t{font_color}stdout:\n---\n{bcolors.OKBLUE}{el.get("stdout")}{font_color}\n---{bcolors.ENDC}')
                if el.get("stderr"):
                    print(f'\t\t{font_color}stderr:\n---\n{bcolors.WARNING}{el.get("stderr")}{font_color}\n---{bcolors.ENDC}')
                print(f'\n')

        if exec_exception:
            print(f'\t{font_color}Execution Exception : {exec_exception}{bcolors.ENDC}')

        # print(f"{bcolors.OKBLUE}{json.dumps(router_response_response_payload_dict, indent=4)}{bcolors.ENDC}")

    elif response_type == 'gateway_info':

        gateway_info = router_response_response_payload_dict.get('data')

        # Version and Gateway Details
        print(f'\n{bcolors.OKBLUE}Gateway Details{bcolors.ENDC}')
        gateway_config = gateway_info.get('gateway-config', {})
        version_info = gateway_config.get('version', {})
        if version_info.get("current"):
            print(f'\t{bcolors.OKGREEN}Version           : {version_info.get("current")}{bcolors.ENDC}')

        # Convert Unix timestamp to readable format
        started_time = gateway_config.get("connection_info", {}).get("started")
        try:
            if started_time:
                started_dt = datetime.fromtimestamp(float(started_time))
                local_tz = datetime.now().astimezone().tzinfo
                started_str = f"{started_dt.strftime('%Y-%m-%d %H:%M:%S')} {local_tz}"
                print(f'\t{bcolors.OKGREEN}Started Time      : {started_str}{bcolors.ENDC}')
        except (ValueError, TypeError):
            pass  # Skip if timestamp is invalid

        if gateway_config.get("ws_log_file"):
            print(f'\t{bcolors.OKGREEN}Logs Location     : {gateway_config.get("ws_log_file")}{bcolors.ENDC}')

        # Environment Info
        machine_env = gateway_info.get('machine', {}).get('environment', {})
        if machine_env and machine_env.get('provider'):
            print(f'\n{bcolors.OKBLUE}Environment Details{bcolors.ENDC}')
            env_color = bcolors.WARNING if machine_env.get('provider') == 'Local/Other' else bcolors.OKGREEN
            print(f'\t{env_color}Provider          : {machine_env.get("provider")}{bcolors.ENDC}')
            if machine_env.get('provider') != 'Local/Other':
                if machine_env.get('account_id'):
                    print(f'\t{env_color}Account           : {machine_env.get("account_id")}{bcolors.ENDC}')
                if machine_env.get('region'):
                    print(f'\t{env_color}Region            : {machine_env.get("region")}{bcolors.ENDC}')
                if machine_env.get('instance_type'):
                    print(f'\t{env_color}Instance Type     : {machine_env.get("instance_type")}{bcolors.ENDC}')

        # Machine Details
        machine = gateway_info.get('machine', {})
        print(f'\n{bcolors.OKBLUE}Machine Details{bcolors.ENDC}')

        if machine.get("hostname"):
            print(f'\t{bcolors.OKGREEN}Hostname          : {machine.get("hostname")}{bcolors.ENDC}')
        if machine.get("ip_address_local") and machine.get("ip_address_local") != "unknown":
            print(f'\t{bcolors.OKGREEN}IP (Local)        : {machine.get("ip_address_local")}{bcolors.ENDC}')
        if machine.get("ip_address_external"):
            print(f'\t{bcolors.OKGREEN}IP (External)     : {machine.get("ip_address_external")}{bcolors.ENDC}')

        os_info = []
        if machine.get("system"): os_info.append(machine.get("system"))
        if machine.get("release"): os_info.append(machine.get("release"))
        if os_info:
            print(f'\t{bcolors.OKGREEN}Operating System  : {" ".join(os_info)}{bcolors.ENDC}')

        memory = machine.get('memory', {})
        if memory.get('free_gb') is not None and memory.get('total_gb') is not None:
            print(
                f'\t{bcolors.OKGREEN}Memory            : {memory.get("free_gb")}GB free / {memory.get("total_gb")}GB total{bcolors.ENDC}')

        # Core Package Versions - Extract from installed packages
        installed_packages = {
            pkg.split('==')[0]: pkg.split('==')[1]
            for pkg in machine.get('installed-python-packages', [])
        }

        core_packages = [
            ('KDNRM', installed_packages.get('kdnrm')),
            ('Keeper GraphSync', installed_packages.get('keeper-dag')),
            ('Discovery Common', installed_packages.get('discovery-common'))
        ]

        # Only print Core Components section if at least one core package is found
        if any(version for _, version in core_packages):
            print(f'\n{bcolors.OKBLUE}Core Components{bcolors.ENDC}')
            for name, version in core_packages:
                if version:  # Only print if version is found
                    print(f'\t{bcolors.OKGREEN}{name:<16} : {version}{bcolors.ENDC}')

        # KSM Details
        print(f'\n{bcolors.OKBLUE}KSM Application Details{bcolors.ENDC}')
        ksm_app = gateway_info.get('ksm', {}).get('app', {})
        warnings_color = bcolors.WARNING if ksm_app.get("warnings") else bcolors.OKGREEN

        if ksm_app.get("title"):
            print(f'\t{bcolors.OKGREEN}Title             : {ksm_app.get("title")}{bcolors.ENDC}')
        if ksm_app.get("records-count") is not None:
            print(f'\t{bcolors.OKGREEN}Records Count     : {ksm_app.get("records-count")}{bcolors.ENDC}')
        if ksm_app.get("folders-count") is not None:
            print(f'\t{bcolors.OKGREEN}Folders Count     : {ksm_app.get("folders-count")}{bcolors.ENDC}')
        if ksm_app.get("expires-on"):
            print(f'\t{bcolors.OKGREEN}Expires On        : {ksm_app.get("expires-on")}{bcolors.ENDC}')
        print(f'\t{warnings_color}Warnings          : {ksm_app.get("warnings") or "None"}{bcolors.ENDC}')

        # Router Details
        print(f'\n{bcolors.OKBLUE}Router Connection{bcolors.ENDC}')
        router_conn = gateway_info.get('router', {}).get('connection', {})
        if router_conn.get("base-url"):
            print(f'\t{bcolors.OKGREEN}URL               : {router_conn.get("base-url")}{bcolors.ENDC}')
        router_status = router_conn.get("status", "UNKNOWN").lower()
        status_color = bcolors.OKGREEN if router_status == "connected" else bcolors.WARNING
        print(f'\t{status_color}Status            : {router_status}{bcolors.ENDC}')

        # PAM Configurations
        print(f'\n{bcolors.OKBLUE}PAM Configurations Accessible to this Gateway{bcolors.ENDC}')
        pam_configs = gateway_info.get('pam_configurations', [])
        if pam_configs:
            for idx, config in enumerate(pam_configs, 1):
                print(f'\t{bcolors.OKGREEN}{idx}. {config}{bcolors.ENDC}')
        else:
            print(f'\t{bcolors.WARNING}No PAM Configurations found{bcolors.ENDC}')

        # Additional details for verbose mode
        if is_verbose:
            print(f'\n{bcolors.OKBLUE}Additional Details{bcolors.ENDC}')
            if machine.get("working-dir"):
                print(f'\t{bcolors.OKGREEN}Working Directory : {machine.get("working-dir")}{bcolors.ENDC}')
            if machine.get("package-dir"):
                print(f'\t{bcolors.OKGREEN}Package Directory: {machine.get("package-dir")}{bcolors.ENDC}')
            if machine.get("executable"):
                print(f'\t{bcolors.OKGREEN}Python Executable: {machine.get("executable")}{bcolors.ENDC}')

            if machine.get('installed-python-packages'):
                print(f'\n{bcolors.OKBLUE}Installed Python Packages{bcolors.ENDC}')
                for package in sorted(machine.get('installed-python-packages', [])):
                    print(f'\t{bcolors.OKGREEN}{package}{bcolors.ENDC}')

    # else:
    #     print(f"{bcolors.OKBLUE}{json.dumps(router_response_response_payload_dict, indent=4)}{bcolors.ENDC}")


def print_configs_from_router(params, router_response):
    configurations = router_response.get('response').get('data')

    table = []
    headers = ['Config Uid',
               'Config Name',
               'Access Record Uid',
               'Has access to record?',
               'Access Record Title',
               'Access Record Type']
    for c in configurations:

        configuration_uid = c.get('configurationUid')
        access_record_uid = c.get('accessRecordUid')
        access_record_title = c.get('accessRecordTitle') if 'accessRecordTitle' in c else '[Gateway has no access]'
        access_record_type = c.get('accessRecordType') if 'accessRecordType' in c else '[Gateway has no access]'

        has_access_to_record = access_record_uid in params.record_cache

        row_color = bcolors.FAIL
        if has_access_to_record:
            row_color = bcolors.OKGREEN

        row = [
            f'{row_color}{configuration_uid}',
            c.get('configurationName'),
            access_record_uid,
            'Yes' if has_access_to_record else 'No',
            access_record_title,
            f'{access_record_type}{bcolors.ENDC}'
        ]

        table.append(row)

    table.sort(key=lambda x: (x[3] or '', x[1].lower()))

    base.dump_report_data(table, headers, fmt='table', filename="", row_number=False, column_width=None)


def encrypt_pwd_complexity(rule_list_dict, record_key_unencrypted):
    # With padding
    # rule_list_json_bytes = api.get_record_data_json_bytes(rule_list_dict)
    # rule_list_json_bytes_padded = api.pad_aes_gcm(rule_list_json_bytes)
    # rule_list_json_bytes_padded_encrypted = crypto.encrypt_aes_v2(rule_list_json_bytes_padded, record_key_unencrypted)

    # Without padding
    rule_list_json = json.dumps(rule_list_dict)
    rule_list_json_bytes = rule_list_json.encode('UTF-8')
    rule_list_json_encrypted = crypto.encrypt_aes_v2(rule_list_json_bytes, record_key_unencrypted)

    return rule_list_json_encrypted



from unittest import TestCase, mock

from data_vault import get_synced_params, VaultEnvironment
from keepercommander.commands import register
from keepercommander.proto import APIRequest_pb2, record_pb2
from keepercommander import utils

vault_env = VaultEnvironment()


class TestRegister(TestCase):
    expected_commands = []

    def setUp(self):
        self.communicate_mock = mock.patch('keepercommander.api.run_command').start()
        self.communicate_mock.side_effect = TestRegister.communicate_success
        self.communicate_mock = mock.patch('keepercommander.api.communicate_rest').start()
        self.communicate_mock.side_effect = TestRegister.communicate_rest_success
        TestRegister.expected_commands.clear()

    def tearDown(self):
        mock.patch.stopall()

    def test_share_record(self):
        params = get_synced_params()

        record_uid = next(iter([x['record_uid'] for x in params.meta_data_cache.values() if x['can_share']]))
        cmd = register.ShareRecordCommand()

        def not_shared(params, record_uids, is_share_admin):
            pass
        self.record_share_mock = mock.patch('keepercommander.api.get_record_shares').start()
        self.record_share_mock.side_effect = not_shared

        TestRegister.expected_commands.extend(['records_share_update'])
        cmd.execute(params, email=['user2@keepersecurity.com'], action='grant', can_share=False, can_edit=True, record=record_uid)
        self.assertEqual(len(TestRegister.expected_commands), 0)

        TestRegister.expected_commands.extend(['records_share_update'])
        cmd.execute(params, email=['user2@keepersecurity.com'], action='owner', can_share=False, can_edit=True, record=record_uid)
        self.assertEqual(len(TestRegister.expected_commands), 0)

        def shared(params, record_uids, is_share_admin):
            return [{
                'shares': {
                    'user_permissions': [
                        {
                            'username': params.user,
                            'owner': True,
                        },
                        {
                            'username': 'user2@keepersecurity.com',
                            'owner': False,
                            'shareable': False,
                            'editable': False
                        }
                    ]
                }
            }]
        self.record_share_mock.side_effect = shared

        TestRegister.expected_commands.extend(['records_share_update'])
        cmd.execute(params, email=['user2@keepersecurity.com'], action='revoke', record=record_uid)
        self.assertEqual(len(TestRegister.expected_commands), 0)

    def test_share_folder(self):
        params = get_synced_params()
        shared_folder_uid = next(iter([x['shared_folder_uid'] for x in params.shared_folder_cache.values()]))
        cmd = register.ShareFolderCommand()

        TestRegister.expected_commands.extend(['shared_folder_update_v3'])
        cmd.execute(params, action='grant', user=['user2@keepersecurity.com'], manage_records=True, manage_users=False, folder=shared_folder_uid)
        self.assertEqual(len(TestRegister.expected_commands), 0)

        TestRegister.expected_commands.extend(['shared_folder_update_v3'])
        cmd.execute(params, action='revoke', user=['user2@keepersecurity.com'], folder=shared_folder_uid)
        self.assertEqual(len(TestRegister.expected_commands), 0)

    @staticmethod
    def record_share_rq_rs(rq):
        status = record_pb2.SharedRecordStatus()
        status.recordUid = rq.recordUid
        status.status = 'success'
        status.username = rq.toUsername
        return status

    @staticmethod
    def communicate_rest_success(params, request, endpoint, **kwargs):
        if 'rs_type' in kwargs:
            rs = kwargs['rs_type']()
        else:
            rs = None

        _, _, command = endpoint.rpartition('/')

        if command == 'get_public_keys':
            for x in request.usernames:
                key_response = APIRequest_pb2.PublicKeyResponse()
                key_response.username = x
                key_response.publicKey = utils.base64_url_decode(vault_env.encoded_public_key)
                rs.keyResponses.append(key_response)
            return rs
        if command == 'records_share_update':
            rs.addSharedRecordStatus.extend((TestRegister.record_share_rq_rs(x) for x in request.addSharedRecord))
            rs.addSharedRecordStatus.extend((TestRegister.record_share_rq_rs(x) for x in request.updateSharedRecord))
            rs.removeSharedRecordStatus.extend((TestRegister.record_share_rq_rs(x) for x in request.removeSharedRecord))

        cmd = TestRegister.expected_commands.pop(0)
        if cmd == command:
            return rs

        raise Exception()

    @staticmethod
    def communicate_success(params, request):
        rs = {
            'result': 'success',
            'result_code': '',
            'message': ''
        }
        raise Exception()



import os
import json

from keepercommander.params import KeeperParams


def read_config_file(params, file_name='config.json'):       # type: (KeeperParams, str) -> None
    params.config_filename = os.path.join(os.path.dirname(__file__), file_name)
    with open(params.config_filename, 'r', encoding='utf-8') as f:
        params.config = json.load(f)
        params.server = params.config['server']
        params.user = params.config['user']
        params.device_private_key = params.config['private_key']
        params.device_token = params.config['device_token']
        params.clone_code = params.config['clone_code']
        params.password = params.config.get('password', None)




import os
from io import RawIOBase, BufferedReader, IOBase, BytesIO

from Cryptodome.Cipher import AES


CHUNK_SIZE = 8 * 1024


def encode_aes_from_stream(stream, encryption_key, aes_mode=AES.MODE_GCM, chunk_size=CHUNK_SIZE):
    """Encrypts using AES from file in chunks"""
    iv = os.urandom(12)
    yield iv
    cipher = AES.new(key=encryption_key, mode=aes_mode, nonce=iv)
    chunk = stream.read(CHUNK_SIZE)
    while chunk:
        yield cipher.encrypt(chunk)
        chunk = stream.read(CHUNK_SIZE)
    yield cipher.digest()


class EncryptionReader(RawIOBase):
    """A RawIOBase reader that encrypts the input stream"""

    def __init__(self, filename_or_stream, key):
        if isinstance(filename_or_stream, IOBase):
            self.stream = filename_or_stream
        elif os.path.isfile(filename_or_stream):
            self.stream = open(filename_or_stream, 'rb')
        else:
            # Fail silently with empty stream
            self.stream = BytesIO()

        self.encryption_generator = encode_aes_from_stream(self.stream, key)
        self.leftover = None

    def readable(self):
        return True

    def readinto(self, b):
        try:
            buf_len = len(b)
            chunk = self.leftover or next(self.encryption_generator)
            output = chunk[:buf_len]
            self.leftover = chunk[buf_len:]
            ret_len = len(output)
            b[:ret_len] = output
            return ret_len
        except StopIteration:
            return 0

    def close(self):
        self.stream.close()

    @classmethod
    def get_buffered_reader(cls, stream, key):
        return BufferedReader(cls(stream, key))



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: DeviceManagement.proto
# Protobuf Python Version: 5.29.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import APIRequest_pb2 as APIRequest__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16\x44\x65viceManagement.proto\x12\x10\x44\x65viceManagement\x1a\x10\x41PIRequest.proto\"\x97\x03\n\x06\x44\x65vice\x12\x1c\n\x14\x65ncryptedDeviceToken\x18\x01 \x01(\x0c\x12\x12\n\ndeviceName\x18\x02 \x01(\t\x12\x16\n\x0e\x64\x65vicePlatform\x18\x03 \x01(\t\x12\x32\n\x0c\x64\x65viceStatus\x18\x04 \x01(\x0e\x32\x1c.Authentication.DeviceStatus\x12.\n\nloginState\x18\x05 \x01(\x0e\x32\x1a.Authentication.LoginState\x12\x15\n\rclientVersion\x18\x06 \x01(\t\x12\x30\n\nclientType\x18\x07 \x01(\x0e\x32\x1c.DeviceManagement.ClientType\x12@\n\x12\x63lientTypeCategory\x18\x08 \x01(\x0e\x32$.DeviceManagement.ClientTypeCategory\x12:\n\x10\x63lientFormFactor\x18\t \x01(\x0e\x32 .Authentication.ClientFormFactor\x12\x18\n\x10lastModifiedTime\x18\n \x01(\x03\"8\n\x0b\x44\x65viceGroup\x12)\n\x07\x64\x65vices\x18\x01 \x03(\x0b\x32\x18.DeviceManagement.Device\"I\n\x12\x44\x65viceUserResponse\x12\x33\n\x0c\x64\x65viceGroups\x18\x01 \x03(\x0b\x32\x1d.DeviceManagement.DeviceGroup\"K\n\x13\x44\x65viceActionRequest\x12\x34\n\x0c\x64\x65viceAction\x18\x01 \x03(\x0b\x32\x1e.DeviceManagement.DeviceAction\"j\n\x0c\x44\x65viceAction\x12<\n\x10\x64\x65viceActionType\x18\x01 \x01(\x0e\x32\".DeviceManagement.DeviceActionType\x12\x1c\n\x14\x65ncryptedDeviceToken\x18\x02 \x03(\x0c\"\x8d\x01\n\x14\x44\x65viceActionResponse\x12@\n\x12\x64\x65viceActionResult\x18\x01 \x03(\x0b\x32$.DeviceManagement.DeviceActionResult\x12\x33\n\x0c\x64\x65viceGroups\x18\x02 \x03(\x0b\x32\x1d.DeviceManagement.DeviceGroup\"\xb2\x01\n\x12\x44\x65viceActionResult\x12<\n\x10\x64\x65viceActionType\x18\x01 \x01(\x0e\x32\".DeviceManagement.DeviceActionType\x12\x1c\n\x14\x65ncryptedDeviceToken\x18\x02 \x03(\x0c\x12@\n\x12\x64\x65viceActionStatus\x18\x03 \x01(\x0e\x32$.DeviceManagement.DeviceActionStatus\"K\n\x13\x44\x65viceRenameRequest\x12\x34\n\x0c\x64\x65viceRename\x18\x01 \x03(\x0b\x32\x1e.DeviceManagement.DeviceRename\"C\n\x0c\x44\x65viceRename\x12\x1c\n\x14\x65ncryptedDeviceToken\x18\x01 \x01(\x0c\x12\x15\n\rdeviceNewName\x18\x02 \x01(\t\"\x8d\x01\n\x14\x44\x65viceRenameResponse\x12@\n\x12\x64\x65viceRenameResult\x18\x01 \x03(\x0b\x32$.DeviceManagement.DeviceRenameResult\x12\x33\n\x0c\x64\x65viceGroups\x18\x02 \x03(\x0b\x32\x1d.DeviceManagement.DeviceGroup\"\x8b\x01\n\x12\x44\x65viceRenameResult\x12\x1c\n\x14\x65ncryptedDeviceToken\x18\x01 \x01(\x0c\x12\x15\n\rdeviceNewName\x18\x02 \x01(\t\x12@\n\x12\x64\x65viceActionStatus\x18\x03 \x01(\x0e\x32$.DeviceManagement.DeviceActionStatus\"/\n\x12\x44\x65viceAdminRequest\x12\x19\n\x11\x65nterpriseUserIds\x18\x01 \x03(\x03\"O\n\x13\x44\x65viceAdminResponse\x12\x38\n\x0e\x64\x65viceUserList\x18\x01 \x03(\x0b\x32 .DeviceManagement.DeviceUserList\"_\n\x0e\x44\x65viceUserList\x12\x18\n\x10\x65nterpriseUserId\x18\x01 \x01(\x03\x12\x33\n\x0c\x64\x65viceGroups\x18\x02 \x03(\x0b\x32\x1d.DeviceManagement.DeviceGroup\"Z\n\x18\x44\x65viceAdminActionRequest\x12>\n\x11\x64\x65viceAdminAction\x18\x01 \x03(\x0b\x32#.DeviceManagement.DeviceAdminAction\"\x89\x01\n\x11\x44\x65viceAdminAction\x12<\n\x10\x64\x65viceActionType\x18\x01 \x01(\x0e\x32\".DeviceManagement.DeviceActionType\x12\x18\n\x10\x65nterpriseUserId\x18\x02 \x01(\x03\x12\x1c\n\x14\x65ncryptedDeviceToken\x18\x03 \x03(\x0c\"h\n\x19\x44\x65viceAdminActionResponse\x12K\n\x18\x64\x65viceAdminActionResults\x18\x01 \x03(\x0b\x32).DeviceManagement.DeviceAdminActionResult\"\xd1\x01\n\x17\x44\x65viceAdminActionResult\x12<\n\x10\x64\x65viceActionType\x18\x01 \x01(\x0e\x32\".DeviceManagement.DeviceActionType\x12\x18\n\x10\x65nterpriseUserId\x18\x02 \x01(\x03\x12\x1c\n\x14\x65ncryptedDeviceToken\x18\x03 \x03(\x0c\x12@\n\x12\x64\x65viceActionStatus\x18\x04 \x01(\x0e\x32$.DeviceManagement.DeviceActionStatus*\xb2\x01\n\x10\x44\x65viceActionType\x12\x0e\n\nDA_INVALID\x10\x00\x12\r\n\tDA_LOGOUT\x10\x01\x12\r\n\tDA_REMOVE\x10\x02\x12\x0b\n\x07\x44\x41_LOCK\x10\x03\x12\r\n\tDA_UNLOCK\x10\x04\x12\x1a\n\x16\x44\x41_DEVICE_ACCOUNT_LOCK\x10\x05\x12\x1c\n\x18\x44\x41_DEVICE_ACCOUNT_UNLOCK\x10\x06\x12\x0b\n\x07\x44\x41_LINK\x10\x07\x12\r\n\tDA_UNLINK\x10\x08*?\n\x12\x44\x65viceActionStatus\x12\x0b\n\x07INVALID\x10\x00\x12\x0b\n\x07SUCCESS\x10\x01\x12\x0f\n\x0bNOT_ALLOWED\x10\x02*\xec\x05\n\nClientType\x12\x08\n\x04NONE\x10\x00\x12\x0b\n\x07\x41NDROID\x10\x01\x12\x0e\n\nBLACKBERRY\x10\x02\x12\x10\n\x0c\x42LACKBERRY10\x10\x03\x12\x0b\n\x07\x44\x45SKTOP\x10\x04\x12\x07\n\x03IOS\x10\x05\x12\x0b\n\x07MAC_APP\x10\x06\x12\x0b\n\x07WEB_APP\x10\x07\x12\x11\n\rWINDOWS_PHONE\x10\x08\x12\x0b\n\x07SURFACE\x10\t\x12\x08\n\x04WIN8\x10\n\x12\x10\n\x0cIE_EXTENSION\x10\x0b\x12\x14\n\x10\x43HROME_EXTENSION\x10\x0c\x12\x15\n\x11\x46IREFOX_EXTENSION\x10\r\x12\x14\n\x10SAFARI_EXTENSION\x10\x0e\x12\n\n\x06\x44OCOMO\x10\x0f\x12\x0b\n\x07UNKNOWN\x10\x10\x12\n\n\x06SERVER\x10\x11\x12\r\n\tCOMMANDER\x10\x12\x12\n\n\x06\x42RIDGE\x10\x13\x12!\n\x1d\x45NTERPRISE_MANAGEMENT_CONSOLE\x10\x14\x12\x12\n\x0e\x45\x44GE_EXTENSION\x10\x15\x12\x10\n\x0cSUPPORT_TOOL\x10\x16\x12\x0f\n\x0bSSO_CONNECT\x10\x17\x12\x14\n\x10\x44\x45SKTOP_ELECTRON\x10\x18\x12\x15\n\x11PASSWORD_IMPORTER\x10\x19\x12\x08\n\x04\x43HAT\x10\x1a\x12\x14\n\x10MAC_APP_ELECTRON\x10\x1b\x12\x0c\n\x08\x43HAT_IOS\x10\x1c\x12\x10\n\x0c\x43HAT_ANDROID\x10\x1d\x12\x10\n\x0c\x43HAT_WINDOWS\x10\x1e\x12\x0c\n\x08\x43HAT_MAC\x10\x1f\x12\x08\n\x04SCIM\x10 \x12\n\n\x06LAMBDA\x10!\x12\x1e\n\x1a\x43ONNECTWISE_CONTROL_HELPER\x10\"\x12\x1a\n\x16\x45NTERPRISE_CLIENT_TOOL\x10#\x12\x16\n\x12SECRETS_MANAGER_JS\x10$\x12\x1a\n\x16SECRETS_MANAGER_PYTHON\x10%\x12\x16\n\x12SECRETS_MANAGER_GO\x10&\x12\x18\n\x14SECRETS_MANAGER_JAVA\x10\'\x12\x17\n\x13SECRETS_MANAGER_NET\x10(*\xcb\x01\n\x12\x43lientTypeCategory\x12\x0c\n\x08\x43\x41T_NONE\x10\x00\x12\r\n\tCAT_ADMIN\x10\x01\x12\x0f\n\x0b\x43\x41T_DESKTOP\x10\x02\x12\x11\n\rCAT_EXTENSION\x10\x03\x12\x0e\n\nCAT_MOBILE\x10\x04\x12\r\n\tCAT_OTHER\x10\x05\x12\x11\n\rCAT_WEB_VAULT\x10\x06\x12\x14\n\x10\x43\x41T_CHAT_DESKTOP\x10\x07\x12\x13\n\x0f\x43\x41T_CHAT_MOBILE\x10\x08\x12\x17\n\x13\x43\x41T_SECRETS_MANAGER\x10\tB,\n\x18\x63om.keepersecurity.protoB\x10\x44\x65viceManagementb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'DeviceManagement_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n\030com.keepersecurity.protoB\020DeviceManagement'
  _globals['_DEVICEACTIONTYPE']._serialized_start=2325
  _globals['_DEVICEACTIONTYPE']._serialized_end=2503
  _globals['_DEVICEACTIONSTATUS']._serialized_start=2505
  _globals['_DEVICEACTIONSTATUS']._serialized_end=2568
  _globals['_CLIENTTYPE']._serialized_start=2571
  _globals['_CLIENTTYPE']._serialized_end=3319
  _globals['_CLIENTTYPECATEGORY']._serialized_start=3322
  _globals['_CLIENTTYPECATEGORY']._serialized_end=3525
  _globals['_DEVICE']._serialized_start=63
  _globals['_DEVICE']._serialized_end=470
  _globals['_DEVICEGROUP']._serialized_start=472
  _globals['_DEVICEGROUP']._serialized_end=528
  _globals['_DEVICEUSERRESPONSE']._serialized_start=530
  _globals['_DEVICEUSERRESPONSE']._serialized_end=603
  _globals['_DEVICEACTIONREQUEST']._serialized_start=605
  _globals['_DEVICEACTIONREQUEST']._serialized_end=680
  _globals['_DEVICEACTION']._serialized_start=682
  _globals['_DEVICEACTION']._serialized_end=788
  _globals['_DEVICEACTIONRESPONSE']._serialized_start=791
  _globals['_DEVICEACTIONRESPONSE']._serialized_end=932
  _globals['_DEVICEACTIONRESULT']._serialized_start=935
  _globals['_DEVICEACTIONRESULT']._serialized_end=1113
  _globals['_DEVICERENAMEREQUEST']._serialized_start=1115
  _globals['_DEVICERENAMEREQUEST']._serialized_end=1190
  _globals['_DEVICERENAME']._serialized_start=1192
  _globals['_DEVICERENAME']._serialized_end=1259
  _globals['_DEVICERENAMERESPONSE']._serialized_start=1262
  _globals['_DEVICERENAMERESPONSE']._serialized_end=1403
  _globals['_DEVICERENAMERESULT']._serialized_start=1406
  _globals['_DEVICERENAMERESULT']._serialized_end=1545
  _globals['_DEVICEADMINREQUEST']._serialized_start=1547
  _globals['_DEVICEADMINREQUEST']._serialized_end=1594
  _globals['_DEVICEADMINRESPONSE']._serialized_start=1596
  _globals['_DEVICEADMINRESPONSE']._serialized_end=1675
  _globals['_DEVICEUSERLIST']._serialized_start=1677
  _globals['_DEVICEUSERLIST']._serialized_end=1772
  _globals['_DEVICEADMINACTIONREQUEST']._serialized_start=1774
  _globals['_DEVICEADMINACTIONREQUEST']._serialized_end=1864
  _globals['_DEVICEADMINACTION']._serialized_start=1867
  _globals['_DEVICEADMINACTION']._serialized_end=2004
  _globals['_DEVICEADMINACTIONRESPONSE']._serialized_start=2006
  _globals['_DEVICEADMINACTIONRESPONSE']._serialized_end=2110
  _globals['_DEVICEADMINACTIONRESULT']._serialized_start=2113
  _globals['_DEVICEADMINACTIONRESULT']._serialized_end=2322
# @@protoc_insertion_point(module_scope)


# -*- coding: utf-8 -*-
#  _  __  
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|            
#
# Keeper Commander 
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#
import json
import ldap3
import logging
import ssl

from ...vault import KeeperRecord
from ...commands.base import RecordMixin

"""Commander Plugin for Active Directory
   Dependencies: 
       pip install ldap3
"""

PasswordChangeHeader = 'Unable to update the password.'

DomainConstraintViolation = 'The value provided for the new password does not meet ' +\
                            'the length, complexity, or history requirements for the domain.'


def rotate(record, new_password):   # type: (KeeperRecord, str) -> bool
    old_password = RecordMixin.get_record_field(record, 'password')
    if not old_password:
        raise ValueError(f'Rotate AD password: Current password is not set.')

    host = RecordMixin.get_record_field(record, 'host')
    if host:
        host, _, port = host.partition(':')
    else:
        port = ''
    if not port:
        port = RecordMixin.get_record_field(record, 'port')
    if port:
        port = int(port)
    else:
        port = None
    if not host:
        raise ValueError(f'Rotate AD password: Domain controller (\"host\") is not set.')

    user_dn = RecordMixin.get_record_field(record, 'cmdr:userdn')
    login = ''
    if not user_dn:
        login = RecordMixin.get_record_field(record, 'login')
        if login:
            if login.lower().startswith('CN='):
                user_dn = login
                login = ''

    if not login and not user_dn:
        raise ValueError(f'Rotate AD password: User login or DN is not set.')

    tls = ldap3.Tls(validate=ssl.CERT_NONE)
    server = ldap3.Server(host=host, port=port, use_ssl=True, tls=tls, connect_timeout=5, get_info=ldap3.ALL)
    with ldap3.Connection(server) as c:
        c.open()

    if user_dn:
        conn = ldap3.Connection(
            server, version=3, auto_bind=ldap3.AUTO_BIND_NONE, authentication=ldap3.SIMPLE,
            client_strategy=ldap3.SYNC, read_only=False, lazy=False,
            user=user_dn, password=old_password)
    else:
        conn = ldap3.Connection(
            server, version=3, auto_bind=ldap3.AUTO_BIND_NONE, authentication=ldap3.NTLM,
            client_strategy=ldap3.SYNC, read_only=False, lazy=False,
            user=login, password=old_password)

    if not conn.bind():
        raise ValueError(f'Rotate AD password: Bind error: {conn.result}')

    if not user_dn:
        domain, _, name = login.partition('\\')
        if not name:
            raise ValueError(f'Rotate AD password: Cannot get User DN')

        request = f'(&(objectClass=user)(sAMAccountName={name}))'
        conn.search(search_base=server.info.naming_contexts[0], search_filter=request,
                    attributes=['distinguishedName'], search_scope=ldap3.SUBTREE)
        if len(conn.entries) == 0:
            raise ValueError(f'Rotate AD password: Cannot get User DN')

        user_dn = conn.entries[0]['distinguishedName'].value

    change_result = conn.extend.microsoft.modify_password(
        user=user_dn, new_password=new_password, old_password=old_password)

    if not change_result:
        error_result = conn.result
        if isinstance(error_result, dict):
            if error_result.get('description', '') == 'constraintViolation':
                logging.info(f'{PasswordChangeHeader} {DomainConstraintViolation}')
            elif 'message' in error_result:
                logging.info(f'{PasswordChangeHeader}: %s', conn.result["message"])
            else:
                logging.info(f'{PasswordChangeHeader}: %s', json.dumps(conn.result))
        else:
            logging.info(f'{PasswordChangeHeader}: %s', str(conn.result))
    conn.unbind()

    return change_result
